import {
  ENTITIES_KEY_PREFIX,
  Latest,
  getClient,
  getClientGroup,
  getClients,
  inspectQueriesDownSchema,
  nanoid,
  readFromHash,
  withRead
} from "./chunk-IKN2MNWU.js";
import {
  SUBQ_PREFIX,
  normalizeTTL,
  test,
  valita_exports
} from "./chunk-MIYN75UK.js";
import {
  assert,
  must,
  unreachable
} from "./chunk-SKR5JO2A.js";

// ../ast-to-zql/src/ast-to-zql.ts
function astToZQL(ast) {
  let code = "";
  if (ast.where) {
    code += transformCondition(ast.where, ".where", /* @__PURE__ */ new Set());
  }
  if (ast.related && ast.related.length > 0) {
    for (const related of ast.related) {
      if (related.hidden) {
        const nestedRelated = related.subquery.related?.[0];
        if (nestedRelated) {
          code += transformRelated(nestedRelated);
        }
      } else {
        code += transformRelated(related);
      }
    }
  }
  if (ast.orderBy && ast.orderBy.length > 0) {
    code += transformOrder(ast.orderBy);
  }
  if (ast.limit !== void 0) {
    code += `.limit(${ast.limit})`;
  }
  if (ast.start) {
    const { row, exclusive } = ast.start;
    code += `.start(${JSON.stringify(row)}${exclusive ? "" : ", { inclusive: true }"})`;
  }
  return code;
}
function transformCondition(condition, prefix, args) {
  switch (condition.type) {
    case "simple":
      return transformSimpleCondition(condition, prefix);
    case "and":
    case "or":
      return transformLogicalCondition(condition, prefix, args);
    case "correlatedSubquery":
      return transformExistsCondition(condition, prefix, args);
    default:
      unreachable(condition);
  }
}
function transformSimpleCondition(condition, prefix) {
  const { left, op, right } = condition;
  const leftCode = transformValuePosition(left);
  const rightCode = transformValuePosition(right);
  if (op === "=") {
    return `${prefix}(${leftCode}, ${rightCode})`;
  }
  return `${prefix}(${leftCode}, '${op}', ${rightCode})`;
}
function transformLogicalCondition(condition, prefix, args) {
  const { type, conditions } = condition;
  if (conditions.length === 1) {
    return transformCondition(conditions[0], prefix, args);
  }
  if (type === "and") {
    const parts = conditions.map((c) => transformCondition(c, prefix, args));
    if (prefix === ".where") {
      return parts.join("");
    }
    args.add("and");
    return "and(" + parts.join(", ") + ")";
  }
  args = /* @__PURE__ */ new Set();
  const conditionsCode = conditions.map((c) => transformCondition(c, "cmp", args)).join(", ");
  args.add("cmp");
  args.add(type);
  const argsCode = [...args].sort().join(", ");
  return `.where(({${argsCode}}) => ${type}(${conditionsCode}))`;
}
function transformExistsCondition(condition, prefix, args) {
  const { related, op } = condition;
  const relationship = extractRelationshipName(related);
  const hasSubQueryProps = related.subquery.where || related.subquery.related && related.subquery.related.length > 0 || related.subquery.orderBy || related.subquery.limit;
  if (op === "EXISTS") {
    if (!hasSubQueryProps) {
      if (prefix === ".where") {
        return `.whereExists('${relationship}')`;
      }
      args.add("exists");
      return `exists('${relationship}')`;
    }
    if (prefix === ".where") {
      return `.whereExists('${relationship}', q => q${astToZQL(
        related.subquery
      )})`;
    }
    prefix;
    args.add("exists");
    return `exists('${relationship}', q => q${astToZQL(related.subquery)})`;
  }
  op;
  if (hasSubQueryProps) {
    if (prefix === ".where") {
      return `.where(({exists, not}) => not(exists('${relationship}', q => q${astToZQL(
        related.subquery
      )})))`;
    }
    prefix;
    args.add("not");
    args.add("exists");
    return `not(exists('${relationship}', q => q${astToZQL(
      related.subquery
    )}))`;
  }
  if (prefix === ".where") {
    return `.where(({exists, not}) => not(exists('${relationship}')))`;
  }
  args.add("not");
  args.add("exists");
  return `not(exists('${relationship}')))`;
}
function extractRelationshipName(related) {
  const alias = must(related.subquery.alias);
  return alias.startsWith(SUBQ_PREFIX) ? alias.substring(SUBQ_PREFIX.length) : alias;
}
function transformRelated(related) {
  const { alias } = related.subquery;
  if (!alias) return "";
  const relationship = alias;
  let code = `.related('${relationship}'`;
  if (related.subquery.where || related.subquery.related && related.subquery.related.length > 0 || related.subquery.orderBy || related.subquery.limit) {
    code += ", q => q" + astToZQL(related.subquery);
  }
  code += ")";
  return code;
}
function transformOrder(orderBy) {
  let code = "";
  for (const [field, direction] of orderBy) {
    code += `.orderBy('${field}', '${direction}')`;
  }
  return code;
}
function transformValuePosition(value) {
  switch (value.type) {
    case "literal":
      return transformLiteral(value);
    case "column":
      return `'${value.name}'`;
    case "static":
      return transformParameter(value);
    default:
      unreachable(value);
  }
}
function transformLiteral(literal) {
  if (literal.value === null) {
    return "null";
  }
  if (Array.isArray(literal.value)) {
    return JSON.stringify(literal.value);
  }
  if (typeof literal.value === "string") {
    return `'${literal.value.replace(/'/g, "\\'")}'`;
  }
  return String(literal.value);
}
function transformParameter(param) {
  const fieldStr = Array.isArray(param.field) ? `[${param.field.map((f) => `'${f}'`).join(", ")}]` : `'${param.field}'`;
  return `authParam(${fieldStr})`;
}

// ../zero-client/src/client/inspector/inspector.ts
async function newInspector(rep, schema, socket) {
  const clientGroupID = await rep.clientGroupID;
  return new Inspector(rep, schema, rep.clientID, clientGroupID, socket);
}
var Inspector = class {
  #rep;
  client;
  clientGroup;
  #schema;
  socket;
  constructor(rep, schema, clientID, clientGroupID, socket) {
    this.#rep = rep;
    this.#schema = schema;
    this.client = new Client(rep, schema, socket, clientID, clientGroupID);
    this.clientGroup = this.client.clientGroup;
    this.socket = socket;
  }
  clients() {
    return withDagRead(
      this.#rep,
      (dagRead) => clients(this.#rep, this.socket, this.#schema, dagRead)
    );
  }
  clientsWithQueries() {
    return withDagRead(
      this.#rep,
      (dagRead) => clientsWithQueries(this.#rep, this.socket, this.#schema, dagRead)
    );
  }
};
function rpc(socket, arg, downSchema) {
  return new Promise((resolve, reject) => {
    const id = nanoid();
    const f = (ev) => {
      const msg = JSON.parse(ev.data);
      if (msg[0] === "inspect") {
        const body = msg[1];
        if (body.id !== id) {
          return;
        }
        const res = test(body, downSchema);
        if (res.ok) {
          resolve(res.value.value);
        } else {
          reject(res.error);
        }
        socket.removeEventListener("message", f);
      }
    };
    socket.addEventListener("message", f);
    socket.send(
      JSON.stringify(["inspect", { ...arg, id }])
    );
  });
}
var Client = class {
  #rep;
  id;
  clientGroup;
  #schema;
  #socket;
  constructor(rep, schema, socket, id, clientGroupID) {
    this.#rep = rep;
    this.#schema = schema;
    this.#socket = socket;
    this.id = id;
    this.clientGroup = new ClientGroup(rep, socket, schema, clientGroupID);
  }
  async queries() {
    const rows = await rpc(
      await this.#socket(),
      { op: "queries", clientID: this.id },
      inspectQueriesDownSchema
    );
    return rows.map((row) => new Query(row, this.#schema));
  }
  map() {
    return withDagRead(this.#rep, async (dagRead) => {
      const tree = await getBTree(dagRead, this.id);
      const map = /* @__PURE__ */ new Map();
      for await (const [key, value] of tree.scan("")) {
        map.set(key, value);
      }
      return map;
    });
  }
  rows(tableName) {
    return withDagRead(this.#rep, async (dagRead) => {
      const prefix = ENTITIES_KEY_PREFIX + tableName;
      const tree = await getBTree(dagRead, this.id);
      const rows = [];
      for await (const [key, value] of tree.scan(prefix)) {
        if (!key.startsWith(prefix)) {
          break;
        }
        rows.push(value);
      }
      return rows;
    });
  }
};
var ClientGroup = class {
  #rep;
  id;
  #schema;
  #socket;
  constructor(rep, socket, schema, id) {
    this.#rep = rep;
    this.#socket = socket;
    this.#schema = schema;
    this.id = id;
  }
  clients() {
    return withDagRead(
      this.#rep,
      (dagRead) => clients(
        this.#rep,
        this.#socket,
        this.#schema,
        dagRead,
        ([_, v]) => v.clientGroupID === this.id
      )
    );
  }
  clientsWithQueries() {
    return withDagRead(
      this.#rep,
      (dagRead) => clientsWithQueries(
        this.#rep,
        this.#socket,
        this.#schema,
        dagRead,
        ([_, v]) => v.clientGroupID === this.id
      )
    );
  }
  async queries() {
    const rows = await rpc(
      await this.#socket(),
      { op: "queries" },
      inspectQueriesDownSchema
    );
    return rows.map((row) => new Query(row, this.#schema));
  }
};
async function withDagRead(rep, f) {
  await rep.refresh();
  await rep.persist();
  return withRead(rep.perdag, f);
}
async function getBTree(dagRead, clientID) {
  const client = await getClient(clientID, dagRead);
  assert(client, `Client not found: ${clientID}`);
  const { clientGroupID } = client;
  const clientGroup = await getClientGroup(clientGroupID, dagRead);
  assert(clientGroup, `Client group not found: ${clientGroupID}`);
  const dbRead = await readFromHash(
    clientGroup.headHash,
    dagRead,
    Latest
  );
  return dbRead.map;
}
async function clients(rep, socket, schema, dagRead, predicate = () => true) {
  const clients2 = await getClients(dagRead);
  return [...clients2.entries()].filter(predicate).map(
    ([clientID, { clientGroupID }]) => new Client(rep, schema, socket, clientID, clientGroupID)
  );
}
async function clientsWithQueries(rep, socket, schema, dagRead, predicate = () => true) {
  const allClients = await clients(rep, socket, schema, dagRead, predicate);
  const clientsWithQueries2 = [];
  await Promise.all(
    allClients.map(async (client) => {
      const queries = await client.queries();
      if (queries.length > 0) {
        clientsWithQueries2.push(client);
      }
    })
  );
  return clientsWithQueries2;
}
var Query = class {
  ast;
  got;
  ttl;
  inactivatedAt;
  rowCount;
  deleted;
  id;
  zql;
  clientID;
  constructor(row, _schema) {
    this.clientID = row.clientID;
    this.id = row.queryID;
    this.inactivatedAt = row.inactivatedAt === null ? null : new Date(row.inactivatedAt);
    this.ttl = normalizeTTL(row.ttl);
    this.ast = row.ast;
    this.got = row.got;
    this.rowCount = row.rowCount;
    this.deleted = row.deleted;
    this.zql = this.ast.table + astToZQL(this.ast);
  }
};
export {
  newInspector
};
//# sourceMappingURL=inspector-2QLL7CGJ.js.map
