import {
  Zero
} from "./chunk-TWNGMKRD.js";
import "./chunk-IKN2MNWU.js";
import {
  DEFAULT_TTL
} from "./chunk-MIYN75UK.js";
import {
  applyChange,
  must
} from "./chunk-SKR5JO2A.js";

// ../zero-solid/src/create-query.ts
import { createMemo, onCleanup } from "solid-js";

// ../shared/src/ref-count.ts
var RefCount = class {
  #map = /* @__PURE__ */ new WeakMap();
  /**
   * Returns true if the value was added.
   */
  inc(value) {
    const rc = this.#map.get(value) ?? 0;
    this.#map.set(value, rc + 1);
    return rc === 0;
  }
  /**
   * Returns true if the value was removed.
   */
  dec(value) {
    const rc = must(this.#map.get(value));
    if (rc === 1) {
      return this.#map.delete(value);
    }
    this.#map.set(value, rc - 1);
    return false;
  }
};

// ../zero-solid/src/solid-view.ts
import {
  createStore,
  produce
} from "solid-js/store";
var complete = { type: "complete" };
var unknown = { type: "unknown" };
var SolidView = class {
  #input;
  #format;
  #onDestroy;
  #state;
  #setState;
  // Optimization: if the store is currently empty we build up
  // the view on a plain old JS object stored at #builderRoot, and return
  // that for the new state on transaction commit.  This avoids building up
  // large views from scratch via solid produce.  The proxy object used by
  // solid produce is slow and in this case we don't care about solid tracking
  // the fine grained changes (everything has changed, it's all new).  For a
  // test case with a view with 3000 rows, each row having 2 children, this
  // optimization reduced #applyChanges time from 743ms to 133ms.
  #builderRoot;
  #pendingChanges = [];
  #updateTTL;
  constructor(input, onTransactionCommit, format, onDestroy, queryComplete, updateTTL) {
    this.#input = input;
    onTransactionCommit(this.#onTransactionCommit);
    this.#format = format;
    this.#onDestroy = onDestroy;
    this.#updateTTL = updateTTL;
    input.setOutput(this);
    const initialRoot = this.#createEmptyRoot();
    this.#applyChangesToRoot(
      input.fetch({}),
      (node) => ({ type: "add", node }),
      initialRoot
    );
    [this.#state, this.#setState] = createStore([
      initialRoot,
      queryComplete === true ? complete : unknown
    ]);
    if (isEmptyRoot(initialRoot)) {
      this.#builderRoot = this.#createEmptyRoot();
    }
    if (queryComplete !== true) {
      void queryComplete.then(() => {
        this.#setState((oldState) => [oldState[0], complete]);
      });
    }
  }
  get data() {
    return this.#state[0][""];
  }
  get resultDetails() {
    return this.#state[1];
  }
  destroy() {
    this.#onDestroy();
  }
  #onTransactionCommit = () => {
    const builderRoot = this.#builderRoot;
    if (builderRoot) {
      if (!isEmptyRoot(builderRoot)) {
        this.#setState((oldState) => [builderRoot, oldState[1]]);
        this.#builderRoot = void 0;
      }
    } else {
      try {
        this.#applyChanges(this.#pendingChanges, (c) => c);
      } finally {
        this.#pendingChanges = [];
      }
    }
  };
  push(change) {
    if (this.#builderRoot) {
      this.#applyChangeToRoot(change, this.#builderRoot);
    } else {
      this.#pendingChanges.push(materializeRelationships(change));
    }
  }
  #applyChanges(changes, mapper) {
    this.#setState(
      produce((draftState) => {
        this.#applyChangesToRoot(changes, mapper, draftState[0]);
        if (isEmptyRoot(draftState[0])) {
          this.#builderRoot = this.#createEmptyRoot();
        }
      })
    );
  }
  #applyChangesToRoot(changes, mapper, root) {
    for (const change of changes) {
      this.#applyChangeToRoot(mapper(change), root);
    }
  }
  #applyChangeToRoot(change, root) {
    applyChange(root, change, this.#input.getSchema(), "", this.#format);
  }
  #createEmptyRoot() {
    return {
      "": this.#format.singular ? void 0 : []
    };
  }
  updateTTL(ttl) {
    this.#updateTTL(ttl);
  }
};
function materializeRelationships(change) {
  switch (change.type) {
    case "add":
      return { type: "add", node: materializeNodeRelationships(change.node) };
    case "remove":
      return { type: "remove", node: materializeNodeRelationships(change.node) };
    case "child":
      return {
        type: "child",
        node: { row: change.node.row },
        child: {
          relationshipName: change.child.relationshipName,
          change: materializeRelationships(change.child.change)
        }
      };
    case "edit":
      return {
        type: "edit",
        node: { row: change.node.row },
        oldNode: { row: change.oldNode.row }
      };
  }
}
function materializeNodeRelationships(node) {
  const relationships = {};
  for (const relationship in node.relationships) {
    const materialized = [];
    for (const n of node.relationships[relationship]()) {
      materialized.push(materializeNodeRelationships(n));
    }
    relationships[relationship] = () => materialized;
  }
  return {
    row: node.row,
    relationships
  };
}
function isEmptyRoot(entry) {
  const data = entry[""];
  return data === void 0 || Array.isArray(data) && data.length === 0;
}
function solidViewFactory(_query, input, format, onDestroy, onTransactionCommit, queryComplete, updateTTL) {
  return new SolidView(
    input,
    onTransactionCommit,
    format,
    onDestroy,
    queryComplete,
    updateTTL
  );
}

// ../zero-solid/src/create-query.ts
function createQuery(querySignal, options) {
  const view = createMemo(() => {
    const query = querySignal();
    const ttl = normalize(options)?.ttl ?? DEFAULT_TTL;
    const view2 = getView(query, ttl);
    onCleanup(() => queueMicrotask(() => releaseView(query, view2)));
    return view2;
  });
  return [() => view().data, () => view().resultDetails];
}
function useQuery(querySignal, options) {
  return createQuery(querySignal, options);
}
var views = /* @__PURE__ */ new WeakMap();
var viewRefCount = new RefCount();
function getView(query, ttl) {
  let view = views.get(query);
  if (!view) {
    view = query.materialize(solidViewFactory, ttl);
    views.set(query, view);
  } else {
    view.updateTTL(ttl);
  }
  viewRefCount.inc(view);
  return view;
}
function releaseView(query, view) {
  if (viewRefCount.dec(view)) {
    views.delete(query);
    view.destroy();
  }
}
function normalize(options) {
  return typeof options === "function" ? options() : options;
}

// ../zero-solid/src/create-zero.ts
import { batch } from "solid-js";
function createZero(options) {
  const opts = {
    ...options,
    batchViewUpdates: batch
  };
  return new Zero(opts);
}
export {
  createQuery,
  createZero,
  useQuery
};
//# sourceMappingURL=solid.js.map
