import { LogContext } from '@rocicorp/logger';
import { tmpdir } from 'node:os';
import path from 'node:path';
import { pid } from 'node:process';
import { assert } from "../../../shared/src/asserts.js";
import { must } from "../../../shared/src/must.js";
import { randInt } from "../../../shared/src/rand.js";
import * as v from "../../../shared/src/valita.js";
import { primaryKeyValueSchema, } from "../../../zero-protocol/src/primary-key.js";
import { bindStaticParameters, buildPipeline, } from "../../../zql/src/builder/builder.js";
import { StaticQuery, staticQuery } from "../../../zql/src/query/static-query.js";
import { Database } from "../../../zqlite/src/db.js";
import { compile, sql } from "../../../zqlite/src/internal/sql.js";
import { fromSQLiteTypes, TableSource, } from "../../../zqlite/src/table-source.js";
import { computeZqlSpecs } from "../db/lite-tables.js";
import { StatementRunner } from "../db/statements.js";
import { DatabaseStorage } from "../services/view-syncer/database-storage.js";
import { mapLiteDataTypeToZqlSchemaValue } from "../types/lite.js";
import { getSchema, reloadPermissionsIfChanged, } from "./load-permissions.js";
import { simplifyCondition } from "../../../zql/src/query/expression.js";
export class WriteAuthorizerImpl {
    #schema;
    #replica;
    #builderDelegate;
    #tableSpecs;
    #tables = new Map();
    #statementRunner;
    #lc;
    #appID;
    #clientGroupID;
    #logConfig;
    #loadedPermissions = null;
    constructor(lc, config, replica, appID, cgID) {
        this.#appID = appID;
        this.#clientGroupID = cgID;
        this.#lc = lc.withContext('class', 'WriteAuthorizerImpl');
        this.#logConfig = config.log;
        this.#schema = getSchema(this.#lc, replica);
        this.#replica = replica;
        const tmpDir = config.storageDBTmpDir ?? tmpdir();
        const writeAuthzStorage = DatabaseStorage.create(lc, path.join(tmpDir, `mutagen-${pid}-${randInt(1000000, 9999999)}`));
        const cgStorage = writeAuthzStorage.createClientGroupStorage(cgID);
        this.#builderDelegate = {
            getSource: name => this.#getSource(name),
            createStorage: () => cgStorage.createStorage(),
            decorateInput: input => input,
            decorateFilterInput: input => input,
        };
        this.#tableSpecs = computeZqlSpecs(this.#lc, replica);
        this.#statementRunner = new StatementRunner(replica);
        this.reloadPermissions();
    }
    reloadPermissions() {
        this.#loadedPermissions = reloadPermissionsIfChanged(this.#lc, this.#statementRunner, this.#appID, this.#loadedPermissions).permissions;
    }
    canPreMutation(authData, ops) {
        for (const op of ops) {
            switch (op.op) {
                case 'insert':
                    // insert does not run pre-mutation checks
                    break;
                case 'update':
                    if (!this.#canUpdate('preMutation', authData, op)) {
                        return false;
                    }
                    break;
                case 'delete':
                    if (!this.#canDelete('preMutation', authData, op)) {
                        return false;
                    }
                    break;
            }
        }
        return true;
    }
    canPostMutation(authData, ops) {
        this.#statementRunner.beginConcurrent();
        try {
            for (const op of ops) {
                const source = this.#getSource(op.tableName);
                switch (op.op) {
                    case 'insert': {
                        source.push({
                            type: 'add',
                            row: op.value,
                        });
                        break;
                    }
                    // TODO(mlaw): what if someone updates the same thing twice?
                    // TODO(aa): It seems like it will just work? source.push()
                    // is going to push the row into the table source, and then the
                    // next requirePreMutationRow will just return the row that was
                    // pushed in.
                    case 'update': {
                        source.push({
                            type: 'edit',
                            oldRow: this.#requirePreMutationRow(op),
                            row: op.value,
                        });
                        break;
                    }
                    case 'delete': {
                        source.push({
                            type: 'remove',
                            row: this.#requirePreMutationRow(op),
                        });
                        break;
                    }
                }
            }
            for (const op of ops) {
                switch (op.op) {
                    case 'insert':
                        if (!this.#canInsert('postMutation', authData, op)) {
                            return false;
                        }
                        break;
                    case 'update':
                        if (!this.#canUpdate('postMutation', authData, op)) {
                            return false;
                        }
                        break;
                    case 'delete':
                        // delete does not run post-mutation checks.
                        break;
                }
            }
        }
        finally {
            this.#statementRunner.rollback();
        }
        return true;
    }
    normalizeOps(ops) {
        return ops.map(op => {
            if (op.op === 'upsert') {
                const preMutationRow = this.#getPreMutationRow(op);
                if (preMutationRow) {
                    return {
                        op: 'update',
                        tableName: op.tableName,
                        primaryKey: op.primaryKey,
                        value: op.value,
                    };
                }
                return {
                    op: 'insert',
                    tableName: op.tableName,
                    primaryKey: op.primaryKey,
                    value: op.value,
                };
            }
            return op;
        });
    }
    #canInsert(phase, authData, op) {
        return this.#timedCanDo(phase, 'insert', authData, op);
    }
    #canUpdate(phase, authData, op) {
        return this.#timedCanDo(phase, 'update', authData, op);
    }
    #canDelete(phase, authData, op) {
        return this.#timedCanDo(phase, 'delete', authData, op);
    }
    #getSource(tableName) {
        let source = this.#tables.get(tableName);
        if (source) {
            return source;
        }
        const tableSpec = this.#tableSpecs.get(tableName);
        if (!tableSpec) {
            throw new Error(`Table ${tableName} not found`);
        }
        const { columns, primaryKey } = tableSpec.tableSpec;
        assert(primaryKey.length);
        source = new TableSource(this.#lc, this.#logConfig, this.#clientGroupID, this.#replica, tableName, Object.fromEntries(Object.entries(columns).map(([name, { dataType }]) => [
            name,
            mapLiteDataTypeToZqlSchemaValue(dataType),
        ])), [primaryKey[0], ...primaryKey.slice(1)]);
        this.#tables.set(tableName, source);
        return source;
    }
    #timedCanDo(phase, action, authData, op) {
        const start = performance.now();
        try {
            const ret = this.#canDo(phase, action, authData, op);
            return ret;
        }
        finally {
            this.#lc.info?.('action:', action, 'duration:', performance.now() - start, 'tableName:', op.tableName, 'primaryKey:', op.primaryKey);
        }
    }
    /**
     * Evaluation order is from static to dynamic, broad to specific.
     * table -> column -> row -> cell.
     *
     * If any step fails, the entire operation is denied.
     *
     * That is, table rules supersede column rules, which supersede row rules,
     *
     * All steps must allow for the operation to be allowed.
     */
    #canDo(phase, action, authData, op) {
        const rules = must(this.#loadedPermissions).permissions?.tables[op.tableName];
        const rowPolicies = rules?.row;
        let rowQuery = staticQuery(this.#schema, op.tableName);
        op.primaryKey.forEach(pk => {
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            rowQuery = rowQuery.where(pk, '=', op.value[pk]);
        });
        let applicableRowPolicy;
        switch (action) {
            case 'insert':
                if (phase === 'postMutation') {
                    applicableRowPolicy = rowPolicies?.insert;
                }
                break;
            case 'update':
                if (phase === 'preMutation') {
                    applicableRowPolicy = rowPolicies?.update?.preMutation;
                }
                else if (phase === 'postMutation') {
                    applicableRowPolicy = rowPolicies?.update?.postMutation;
                }
                break;
            case 'delete':
                if (phase === 'preMutation') {
                    applicableRowPolicy = rowPolicies?.delete;
                }
                break;
        }
        const cellPolicies = rules?.cell;
        const applicableCellPolicies = [];
        if (cellPolicies) {
            for (const [column, policy] of Object.entries(cellPolicies)) {
                if (action === 'update' && op.value[column] === undefined) {
                    // If the cell is not being updated, we do not need to check
                    // the cell rules.
                    continue;
                }
                switch (action) {
                    case 'insert':
                        if (policy.insert && phase === 'postMutation') {
                            applicableCellPolicies.push(policy.insert);
                        }
                        break;
                    case 'update':
                        if (phase === 'preMutation' && policy.update?.preMutation) {
                            applicableCellPolicies.push(policy.update.preMutation);
                        }
                        if (phase === 'postMutation' && policy.update?.postMutation) {
                            applicableCellPolicies.push(policy.update.postMutation);
                        }
                        break;
                    case 'delete':
                        if (policy.delete && phase === 'preMutation') {
                            applicableCellPolicies.push(policy.delete);
                        }
                        break;
                }
            }
        }
        if (!this.#passesPolicyGroup(applicableRowPolicy, applicableCellPolicies, authData, rowQuery)) {
            this.#lc.warn?.(`Permission check failed for ${JSON.stringify(op)}, action ${action}, phase ${phase}, authData: ${JSON.stringify(authData)}, rowPolicies: ${JSON.stringify(applicableRowPolicy)}, cellPolicies: ${JSON.stringify(applicableCellPolicies)}`);
            return false;
        }
        return true;
    }
    #getPreMutationRow(op) {
        const { value } = op;
        const conditions = [];
        const values = [];
        for (const pk of op.primaryKey) {
            conditions.push(sql `${sql.ident(pk)}=?`);
            values.push(v.parse(value[pk], primaryKeyValueSchema));
        }
        const spec = this.#tableSpecs.get(op.tableName);
        if (spec === undefined) {
            throw new Error(`Table ${op.tableName} not found`);
        }
        const ret = this.#statementRunner.get(compile(sql `SELECT ${sql.join(Object.keys(spec.zqlSpec).map(c => sql.ident(c)), sql `,`)} FROM ${sql.ident(op.tableName)} WHERE ${sql.join(conditions, sql ` AND `)}`), ...values);
        if (ret === undefined) {
            return ret;
        }
        return fromSQLiteTypes(spec.zqlSpec, ret);
    }
    #requirePreMutationRow(op) {
        const ret = this.#getPreMutationRow(op);
        assert(ret !== undefined, () => `Pre-mutation row not found for ${JSON.stringify(op.value)}`);
        return ret;
    }
    #passesPolicyGroup(applicableRowPolicy, applicableCellPolicies, authData, rowQuery) {
        if (!this.#passesPolicy(applicableRowPolicy, authData, rowQuery)) {
            return false;
        }
        for (const policy of applicableCellPolicies) {
            if (!this.#passesPolicy(policy, authData, rowQuery)) {
                return false;
            }
        }
        return true;
    }
    /**
     * Defaults to *false* if the policy is empty. At least one rule has to pass
     * for the policy to pass.
     */
    #passesPolicy(policy, authData, rowQuery) {
        if (policy === undefined) {
            return false;
        }
        if (policy.length === 0) {
            return false;
        }
        let rowQueryAst = rowQuery.ast;
        rowQueryAst = bindStaticParameters({
            ...rowQueryAst,
            where: updateWhere(rowQueryAst.where, policy),
        }, {
            authData: authData,
            preMutationRow: undefined,
        });
        const input = buildPipeline(rowQueryAst, this.#builderDelegate);
        try {
            const res = input.fetch({});
            for (const _ of res) {
                // if any row is returned at all, the
                // rule passes.
                return true;
            }
        }
        finally {
            input.destroy();
        }
        // no rows returned by any rules? The policy fails.
        return false;
    }
}
function updateWhere(where, policy) {
    assert(where, 'A where condition must exist for RowQuery');
    return simplifyCondition({
        type: 'and',
        conditions: [
            where,
            {
                type: 'or',
                conditions: policy.map(([action, rule]) => {
                    assert(action);
                    return rule;
                }),
            },
        ],
    });
}
//# sourceMappingURL=write-authorizer.js.map