import { tmpdir } from 'node:os';
import path from 'node:path';
import { pid } from 'node:process';
import { assert } from "../../../shared/src/asserts.js";
import { must } from "../../../shared/src/must.js";
import { randInt } from "../../../shared/src/rand.js";
import * as v from "../../../shared/src/valita.js";
import { assertNormalized } from "../config/normalize.js";
import { getZeroConfig } from "../config/zero-config.js";
import { warmupConnections } from "../db/warmup.js";
import { exitAfter, runUntilKilled } from "../services/life-cycle.js";
import { MutagenService } from "../services/mutagen/mutagen.js";
import { PusherService } from "../services/mutagen/pusher.js";
import { DatabaseStorage } from "../services/view-syncer/database-storage.js";
import { DrainCoordinator } from "../services/view-syncer/drain-coordinator.js";
import { PipelineDriver } from "../services/view-syncer/pipeline-driver.js";
import { Snapshotter } from "../services/view-syncer/snapshotter.js";
import { ViewSyncerService } from "../services/view-syncer/view-syncer.js";
import { pgClient } from "../types/pg.js";
import { parentWorker, singleProcessMode, } from "../types/processes.js";
import { getShardID } from "../types/shards.js";
import { Subscription } from "../types/subscription.js";
import { replicaFileModeSchema, replicaFileName } from "../workers/replicator.js";
import { Syncer } from "../workers/syncer.js";
import { createLogContext } from "./logging.js";
import { startOtelAuto } from "./otel-start.js";
function randomID() {
    return randInt(1, Number.MAX_SAFE_INTEGER).toString(36);
}
export default function runWorker(parent, env, ...args) {
    const config = getZeroConfig(env, args.slice(1));
    assertNormalized(config);
    startOtelAuto();
    const lc = createLogContext(config, { worker: 'syncer' });
    assert(args.length > 0, `replicator mode not specified`);
    const fileMode = v.parse(args[0], replicaFileModeSchema);
    const { cvr, upstream } = config;
    assert(cvr.maxConnsPerWorker);
    assert(upstream.maxConnsPerWorker);
    const replicaFile = replicaFileName(config.replica.file, fileMode);
    lc.debug?.(`running view-syncer on ${replicaFile}`);
    const cvrDB = pgClient(lc, cvr.db, {
        max: cvr.maxConnsPerWorker,
        connection: { ['application_name']: `zero-sync-worker-${pid}-cvr` },
    });
    const upstreamDB = pgClient(lc, upstream.db, {
        max: upstream.maxConnsPerWorker,
        connection: { ['application_name']: `zero-sync-worker-${pid}-upstream` },
    });
    const dbWarmup = Promise.allSettled([
        warmupConnections(lc, cvrDB, 'cvr'),
        warmupConnections(lc, upstreamDB, 'upstream'),
    ]);
    const tmpDir = config.storageDBTmpDir ?? tmpdir();
    const operatorStorage = DatabaseStorage.create(lc, path.join(tmpDir, `sync-worker-${pid}-${randInt(1000000, 9999999)}`));
    const shard = getShardID(config);
    const viewSyncerFactory = (id, sub, drainCoordinator) => {
        const logger = lc
            .withContext('component', 'view-syncer')
            .withContext('clientGroupID', id)
            .withContext('instance', randomID());
        lc.debug?.(`creating view syncer`);
        return new ViewSyncerService(logger, shard, config.taskID, id, cvrDB, new PipelineDriver(logger, config.log, new Snapshotter(logger, replicaFile, shard), shard, operatorStorage.createClientGroupStorage(id), id), sub, drainCoordinator, config.log.slowHydrateThreshold, undefined, config.targetClientRowCount);
    };
    const mutagenFactory = (id) => new MutagenService(lc.withContext('component', 'mutagen').withContext('clientGroupID', id), shard, id, upstreamDB, config);
    const pusherFactory = config.push.url === undefined
        ? undefined
        : (id) => new PusherService(config, lc.withContext('clientGroupID', id), id, must(config.push.url), config.push.apiKey);
    const syncer = new Syncer(lc, config, viewSyncerFactory, mutagenFactory, pusherFactory, parent);
    void dbWarmup.then(() => parent.send(['ready', { ready: true }]));
    return runUntilKilled(lc, parent, syncer);
}
// fork()
if (!singleProcessMode()) {
    void exitAfter(() => runWorker(must(parentWorker), process.env, ...process.argv.slice(2)));
}
//# sourceMappingURL=syncer.js.map