type Window = {
    start: number;
    count: number;
};
/**
 * The limiter is a sliding window rate limiter that allows a maximum
 * number of mutations per window.
 *
 * The window is divided into two parts: the prev window and the next window.
 * The prev window is the part of the window that has already passed, and the
 * next window is the part of the window that is still in the future.
 *
 * The limiter keeps track of the number of mutations in each window. When the
 * current time moves completely past the next window, the limiter creates new
 * windows (this only happens if the limiter is not called for a long time).
 * When the current time moves completely into the next window, the
 * limiter rotates the windows.
 *
 * Case 1: need new windows
 * |----|----|
 *             |----| (sliding window)
 *
 * Case 2: rotate windows
 * |----|----|
 *      |----| (sliding window)
 *
 * or
 *
 * |----|----|
 *         |----| (sliding window)
 *
 * The leading edge of the sliding window is current time.
 *
 * The limiter increments the count for the prev window if the current time is
 * in the prior window. Otherwise, it increments the count for the next
 * window.
 *
 * The limiter computes the total mutations by: taking the number of mutation in the next window and
 * adding the number of mutation from the prev window, weighted by
 * the fraction of the window that has passed. The total mutations must be less
 * than the maximum number of mutations allowed per window.
 *
 * |----|----|
 *    |----| (sliding window)
 * ^-- the sliding window covers 25% of the prior window. Only 25% of that window's count counts.
 * The entirety of the next window's count counts.
 */
export declare class SlidingWindowLimiter {
    #private;
    constructor(windowSizeMs: number, maxMutations: number);
    canDo(): boolean;
    totalCallsForTime(now: number): number;
    get priorWindow(): Window;
    get nextWindow(): Window;
}
export {};
//# sourceMappingURL=sliding-window-limiter.d.ts.map