import type { LogContext } from '@rocicorp/logger';
import type { UserPushParams } from '../../../../zero-protocol/src/connect.ts';
import type { Downstream } from '../../../../zero-protocol/src/down.ts';
import { type PushBody } from '../../../../zero-protocol/src/push.ts';
import { type ZeroConfig } from '../../config/zero-config.ts';
import type { Source } from '../../types/streams.ts';
import { Subscription } from '../../types/subscription.ts';
import type { HandlerResult, StreamResult } from '../../workers/connection.ts';
import type { Service } from '../service.ts';
export interface Pusher {
    readonly pushURL: string | undefined;
    enqueuePush(clientID: string, push: PushBody, jwt: string | undefined): HandlerResult;
    initConnection(clientID: string, wsID: string, userPushParams: UserPushParams | undefined): Source<Downstream>;
}
type Config = Pick<ZeroConfig, 'app' | 'shard'>;
/**
 * Receives push messages from zero-client and forwards
 * them the the user's API server.
 *
 * If the user's API server is taking too long to process
 * the push, the PusherService will add the push to a queue
 * and send pushes in bulk the next time the user's API server
 * is available.
 *
 * - One PusherService exists per client group.
 * - Mutations for a given client are always sent in-order
 * - Mutations for different clients in the same group may be interleaved
 */
export declare class PusherService implements Service, Pusher {
    #private;
    readonly id: string;
    constructor(config: Config, lc: LogContext, clientGroupID: string, pushURL: string, apiKey: string | undefined);
    get pushURL(): string | undefined;
    initConnection(clientID: string, wsID: string, userPushParams: UserPushParams | undefined): Subscription<["deleteClients", {
        readonly clientIDs?: readonly string[] | undefined;
        readonly clientGroupIDs?: readonly string[] | undefined;
    }] | ["connected", {
        timestamp?: number | undefined;
        wsid: string;
    }] | ["pushResponse", {
        mutationIDs?: {
            id: number;
            clientID: string;
        }[] | undefined;
        error: "unsupportedPushVersion";
    } | {
        mutationIDs?: {
            id: number;
            clientID: string;
        }[] | undefined;
        error: "unsupportedSchemaVersion";
    } | {
        mutationIDs?: {
            id: number;
            clientID: string;
        }[] | undefined;
        error: "http";
        status: number;
        details: string;
    } | {
        mutationIDs?: {
            id: number;
            clientID: string;
        }[] | undefined;
        error: "zeroPusher";
        details: string;
    } | {
        mutations: {
            id: {
                id: number;
                clientID: string;
            };
            result: {
                details?: import("../../../../shared/src/json.ts").ReadonlyJSONValue | undefined;
                error: "app";
            } | {
                details?: import("../../../../shared/src/json.ts").ReadonlyJSONValue | undefined;
                error: "oooMutation" | "alreadyProcessed";
            } | {
                data?: import("../../../../shared/src/json.ts").ReadonlyJSONValue | undefined;
            };
        }[];
    }] | import("../../../../zero-protocol/src/error.ts").ErrorMessage | ["inspect", {
        op: "queries";
        id: string;
        value: {
            clientID: string;
            queryID: string;
            ast: import("../../../../zero-protocol/src/ast.ts").AST;
            got: boolean;
            deleted: boolean;
            ttl: number;
            inactivatedAt: number | null;
            rowCount: number;
        }[];
    }] | ["pokeStart", {
        schemaVersions?: {
            minSupportedVersion: number;
            maxSupportedVersion: number;
        } | undefined;
        timestamp?: number | undefined;
        pokeID: string;
        baseCookie: string | null;
    }] | ["pokePart", {
        lastMutationIDChanges?: Record<string, number> | undefined;
        desiredQueriesPatches?: Record<string, ({
            ttl?: number | undefined;
            op: "put";
            hash: string;
            ast: import("../../../../zero-protocol/src/ast.ts").AST;
        } | {
            op: "del";
            hash: string;
        } | {
            op: "clear";
        })[]> | undefined;
        gotQueriesPatch?: ({
            ttl?: number | undefined;
            op: "put";
            hash: string;
            ast: import("../../../../zero-protocol/src/ast.ts").AST;
        } | {
            op: "del";
            hash: string;
        } | {
            op: "clear";
        })[] | undefined;
        rowsPatch?: ({
            op: "put";
            tableName: string;
            value: Readonly<Record<string, import("../../../../shared/src/json.ts").ReadonlyJSONValue | undefined>>;
        } | {
            merge?: import("../../../../shared/src/json.ts").ReadonlyJSONObject | undefined;
            constrain?: string[] | undefined;
            op: "update";
            tableName: string;
            id: Readonly<Record<string, string | number | boolean>>;
        } | {
            op: "del";
            tableName: string;
            id: Readonly<Record<string, string | number | boolean>>;
        } | {
            op: "clear";
        })[] | undefined;
        pokeID: string;
    }] | ["pokeEnd", {
        cancel?: boolean | undefined;
        pokeID: string;
        cookie: string;
    }] | ["pong", {}] | ["pull", {
        cookie: string;
        requestID: string;
        lastMutationIDChanges: Record<string, number>;
    }], ["deleteClients", {
        readonly clientIDs?: readonly string[] | undefined;
        readonly clientGroupIDs?: readonly string[] | undefined;
    }] | ["connected", {
        timestamp?: number | undefined;
        wsid: string;
    }] | ["pushResponse", {
        mutationIDs?: {
            id: number;
            clientID: string;
        }[] | undefined;
        error: "unsupportedPushVersion";
    } | {
        mutationIDs?: {
            id: number;
            clientID: string;
        }[] | undefined;
        error: "unsupportedSchemaVersion";
    } | {
        mutationIDs?: {
            id: number;
            clientID: string;
        }[] | undefined;
        error: "http";
        status: number;
        details: string;
    } | {
        mutationIDs?: {
            id: number;
            clientID: string;
        }[] | undefined;
        error: "zeroPusher";
        details: string;
    } | {
        mutations: {
            id: {
                id: number;
                clientID: string;
            };
            result: {
                details?: import("../../../../shared/src/json.ts").ReadonlyJSONValue | undefined;
                error: "app";
            } | {
                details?: import("../../../../shared/src/json.ts").ReadonlyJSONValue | undefined;
                error: "oooMutation" | "alreadyProcessed";
            } | {
                data?: import("../../../../shared/src/json.ts").ReadonlyJSONValue | undefined;
            };
        }[];
    }] | import("../../../../zero-protocol/src/error.ts").ErrorMessage | ["inspect", {
        op: "queries";
        id: string;
        value: {
            clientID: string;
            queryID: string;
            ast: import("../../../../zero-protocol/src/ast.ts").AST;
            got: boolean;
            deleted: boolean;
            ttl: number;
            inactivatedAt: number | null;
            rowCount: number;
        }[];
    }] | ["pokeStart", {
        schemaVersions?: {
            minSupportedVersion: number;
            maxSupportedVersion: number;
        } | undefined;
        timestamp?: number | undefined;
        pokeID: string;
        baseCookie: string | null;
    }] | ["pokePart", {
        lastMutationIDChanges?: Record<string, number> | undefined;
        desiredQueriesPatches?: Record<string, ({
            ttl?: number | undefined;
            op: "put";
            hash: string;
            ast: import("../../../../zero-protocol/src/ast.ts").AST;
        } | {
            op: "del";
            hash: string;
        } | {
            op: "clear";
        })[]> | undefined;
        gotQueriesPatch?: ({
            ttl?: number | undefined;
            op: "put";
            hash: string;
            ast: import("../../../../zero-protocol/src/ast.ts").AST;
        } | {
            op: "del";
            hash: string;
        } | {
            op: "clear";
        })[] | undefined;
        rowsPatch?: ({
            op: "put";
            tableName: string;
            value: Readonly<Record<string, import("../../../../shared/src/json.ts").ReadonlyJSONValue | undefined>>;
        } | {
            merge?: import("../../../../shared/src/json.ts").ReadonlyJSONObject | undefined;
            constrain?: string[] | undefined;
            op: "update";
            tableName: string;
            id: Readonly<Record<string, string | number | boolean>>;
        } | {
            op: "del";
            tableName: string;
            id: Readonly<Record<string, string | number | boolean>>;
        } | {
            op: "clear";
        })[] | undefined;
        pokeID: string;
    }] | ["pokeEnd", {
        cancel?: boolean | undefined;
        pokeID: string;
        cookie: string;
    }] | ["pong", {}] | ["pull", {
        cookie: string;
        requestID: string;
        lastMutationIDChanges: Record<string, number>;
    }]>;
    enqueuePush(clientID: string, push: PushBody, jwt: string | undefined): Exclude<HandlerResult, StreamResult>;
    run(): Promise<void>;
    stop(): Promise<void>;
}
type PusherEntry = {
    push: PushBody;
    jwt: string | undefined;
    clientID: string;
};
type PusherEntryOrStop = PusherEntry | 'stop';
/**
 * Pushes for different clientIDs could theoretically be interleaved.
 *
 * In order to do efficient batching to the user's API server,
 * we collect all pushes for the same clientID into a single push.
 */
export declare function combinePushes(entries: readonly (PusherEntryOrStop | undefined)[]): [PusherEntry[], boolean];
export {};
//# sourceMappingURL=pusher.d.ts.map