import { groupBy } from "../../../../shared/src/arrays.js";
import { assert } from "../../../../shared/src/asserts.js";
import { must } from "../../../../shared/src/must.js";
import { Queue } from "../../../../shared/src/queue.js";
import * as v from "../../../../shared/src/valita.js";
import { ErrorKind } from "../../../../zero-protocol/src/error-kind.js";
import { pushResponseSchema, } from "../../../../zero-protocol/src/push.js";
import {} from "../../config/zero-config.js";
import { ErrorForClient } from "../../types/error-for-client.js";
import { upstreamSchema } from "../../types/shards.js";
import { Subscription } from "../../types/subscription.js";
import instruments from "../../observability/view-syncer-instruments.js";
const reservedParams = ['schema', 'appID'];
/**
 * Receives push messages from zero-client and forwards
 * them the the user's API server.
 *
 * If the user's API server is taking too long to process
 * the push, the PusherService will add the push to a queue
 * and send pushes in bulk the next time the user's API server
 * is available.
 *
 * - One PusherService exists per client group.
 * - Mutations for a given client are always sent in-order
 * - Mutations for different clients in the same group may be interleaved
 */
export class PusherService {
    id;
    #pusher;
    #queue;
    #stopped;
    constructor(config, lc, clientGroupID, pushURL, apiKey) {
        this.#queue = new Queue();
        this.#pusher = new PushWorker(config, lc, pushURL, apiKey, this.#queue);
        this.id = clientGroupID;
    }
    get pushURL() {
        return this.#pusher.pushURL;
    }
    initConnection(clientID, wsID, userPushParams) {
        return this.#pusher.initConnection(clientID, wsID, userPushParams);
    }
    enqueuePush(clientID, push, jwt) {
        this.#queue.enqueue({ push, jwt, clientID });
        return {
            type: 'ok',
        };
    }
    run() {
        this.#stopped = this.#pusher.run();
        return this.#stopped;
    }
    stop() {
        this.#queue.enqueue('stop');
        return must(this.#stopped, 'Stop was called before `run`');
    }
}
/**
 * Awaits items in the queue then drains and sends them all
 * to the user's API server.
 */
class PushWorker {
    #pushURL;
    #apiKey;
    #queue;
    #lc;
    #config;
    #clients;
    constructor(config, lc, pushURL, apiKey, queue) {
        this.#pushURL = pushURL;
        this.#apiKey = apiKey;
        this.#queue = queue;
        this.#lc = lc.withContext('component', 'pusher');
        this.#config = config;
        this.#clients = new Map();
    }
    get pushURL() {
        return this.#pushURL;
    }
    /**
     * Returns a new downstream stream if the clientID,wsID pair has not been seen before.
     * If a clientID already exists with a different wsID, that client's downstream is cancelled.
     */
    initConnection(clientID, wsID, userParams) {
        const existing = this.#clients.get(clientID);
        if (existing && existing.wsID === wsID) {
            // already initialized for this socket
            throw new Error('Connection was already initialized');
        }
        // client is back on a new connection
        if (existing) {
            existing.downstream.cancel();
        }
        const downstream = Subscription.create({
            cleanup: () => {
                this.#clients.delete(clientID);
            },
        });
        this.#clients.set(clientID, { wsID, downstream, userParams });
        return downstream;
    }
    async run() {
        for (;;) {
            const task = await this.#queue.dequeue();
            const rest = this.#queue.drain();
            const [pushes, terminate] = combinePushes([task, ...rest]);
            for (const push of pushes) {
                const response = await this.#processPush(push);
                await this.#fanOutResponses(response);
            }
            if (terminate) {
                break;
            }
        }
    }
    /**
     * The pusher can end up combining many push requests, from the client group, into a single request
     * to the API server.
     *
     * In that case, many different clients will have their mutations present in the
     * PushResponse.
     *
     * Each client is on a different websocket connection though, so we need to fan out the response
     * to all the clients that were part of the push.
     */
    async #fanOutResponses(response) {
        const responses = [];
        const connectionTerminations = [];
        if ('error' in response) {
            this.#lc.warn?.('The server behind ZERO_PUSH_URL returned a push error.', response);
            const groupedMutationIDs = groupBy(response.mutationIDs ?? [], m => m.clientID);
            for (const [clientID, mutationIDs] of groupedMutationIDs) {
                const client = this.#clients.get(clientID);
                if (!client) {
                    continue;
                }
                // We do not resolve mutations on the client if the push fails
                // as those mutations will be retried.
                if (response.error === 'unsupportedPushVersion' ||
                    response.error === 'unsupportedSchemaVersion') {
                    client.downstream.fail(new ErrorForClient({
                        kind: ErrorKind.InvalidPush,
                        message: response.error,
                    }));
                }
                else if (response.error === 'forClient') {
                    client.downstream.fail(response.cause);
                }
                else {
                    responses.push(client.downstream.push([
                        'pushResponse',
                        {
                            ...response,
                            mutationIDs,
                        },
                    ]).result);
                }
            }
        }
        else {
            const groupedMutations = groupBy(response.mutations, m => m.id.clientID);
            for (const [clientID, mutations] of groupedMutations) {
                const client = this.#clients.get(clientID);
                if (!client) {
                    continue;
                }
                let failure;
                let i = 0;
                for (; i < mutations.length; i++) {
                    const m = mutations[i];
                    if ('error' in m.result) {
                        this.#lc.warn?.('The server behind ZERO_PUSH_URL returned a mutation error.', m.result);
                    }
                    if ('error' in m.result && m.result.error === 'oooMutation') {
                        failure = new ErrorForClient({
                            kind: ErrorKind.InvalidPush,
                            message: 'mutation was out of order',
                        });
                        break;
                    }
                }
                if (failure && i < mutations.length - 1) {
                    this.#lc.error?.('push-response contains mutations after a mutation which should fatal the connection');
                }
                // We do not resolve the mutation on the client if it
                // fails for a reason that will cause it to be retried.
                const successes = failure ? mutations.slice(0, i) : mutations;
                if (successes.length > 0) {
                    responses.push(client.downstream.push(['pushResponse', { mutations: successes }])
                        .result);
                }
                if (failure) {
                    connectionTerminations.push(() => client.downstream.fail(failure));
                }
            }
        }
        try {
            await Promise.allSettled(responses);
        }
        finally {
            connectionTerminations.forEach(cb => cb());
        }
    }
    async #processPush(entry) {
        instruments.counters.customMutations.add(entry.push.mutations.length, {
            clientGroupID: entry.push.clientGroupID,
        });
        instruments.counters.pushes.add(1, {
            clientGroupID: entry.push.clientGroupID,
        });
        const headers = {
            'Content-Type': 'application/json',
        };
        if (this.#apiKey) {
            headers['X-Api-Key'] = this.#apiKey;
        }
        if (entry.jwt) {
            headers['Authorization'] = `Bearer ${entry.jwt}`;
        }
        try {
            const params = new URLSearchParams(this.#pushURL.split('?')[1]);
            for (const [key, value] of Object.entries(this.#clients.get(entry.clientID)?.userParams?.queryParams ?? {})) {
                params.append(key, value);
            }
            for (const reserved of reservedParams) {
                assert(!params.has(reserved), `The push URL cannot contain the reserved query param "${reserved}"`);
            }
            params.append('schema', upstreamSchema({
                appID: this.#config.app.id,
                shardNum: this.#config.shard.num,
            }));
            params.append('appID', this.#config.app.id);
            this.#lc.debug?.('pushing custom mutators to', this.#pushURL, 'with params', params.toString(), 'and body', JSON.stringify(entry.push));
            const response = await fetch(`${this.#pushURL}?${params.toString()}`, {
                method: 'POST',
                headers,
                body: JSON.stringify(entry.push),
            });
            if (!response.ok) {
                const details = await response.text();
                // Zero currently handles all auth errors this way (throws ErrorForClient).
                // Continue doing that until we have an `onError` callback exposed on the top level Zero instance.
                // This:
                // 1. Keeps the API the same for those migrating to custom mutators from CRUD
                // 2. Ensures we only churn the API once, when we have `onError` available.
                //
                // When switching to `onError`, we should stop disconnecting the websocket
                // on auth errors and instead let the token be updated
                // on the existing WS connection. This will give us the chance to skip
                // re-hydrating queries that do not use the modified fields of the token.
                if (response.status === 401) {
                    throw new ErrorForClient({
                        kind: ErrorKind.AuthInvalidated,
                        message: details,
                    });
                }
                return {
                    error: 'http',
                    status: response.status,
                    details,
                    mutationIDs: entry.push.mutations.map(m => ({
                        id: m.id,
                        clientID: m.clientID,
                    })),
                };
            }
            const json = await response.json();
            try {
                return v.parse(json, pushResponseSchema);
            }
            catch (e) {
                this.#lc.error?.('failed to parse push response', JSON.stringify(json));
                throw e;
            }
        }
        catch (e) {
            const mutationIDs = entry.push.mutations.map(m => ({
                id: m.id,
                clientID: m.clientID,
            }));
            this.#lc.error?.('failed to push', e);
            if (e instanceof ErrorForClient) {
                return {
                    error: 'forClient',
                    cause: e,
                    mutationIDs,
                };
            }
            // We do not kill the pusher on error.
            // If the user's API server is down, the mutations will never be acknowledged
            // and the client will eventually retry.
            return {
                error: 'zeroPusher',
                details: String(e),
                mutationIDs,
            };
        }
    }
}
/**
 * Pushes for different clientIDs could theoretically be interleaved.
 *
 * In order to do efficient batching to the user's API server,
 * we collect all pushes for the same clientID into a single push.
 */
export function combinePushes(entries) {
    const pushesByClientID = new Map();
    function collect() {
        const ret = [];
        for (const entries of pushesByClientID.values()) {
            const composite = {
                ...entries[0],
                push: {
                    ...entries[0].push,
                    mutations: [],
                },
            };
            ret.push(composite);
            for (const entry of entries) {
                assertAreCompatiblePushes(composite, entry);
                composite.push.mutations.push(...entry.push.mutations);
            }
        }
        return ret;
    }
    for (const entry of entries) {
        if (entry === 'stop' || entry === undefined) {
            return [collect(), true];
        }
        const { clientID } = entry;
        const existing = pushesByClientID.get(clientID);
        if (existing) {
            existing.push(entry);
        }
        else {
            pushesByClientID.set(clientID, [entry]);
        }
    }
    return [collect(), false];
}
// These invariants should always be true for a given clientID.
// If they are not, we have a bug in the code somewhere.
function assertAreCompatiblePushes(left, right) {
    assert(left.clientID === right.clientID, 'clientID must be the same for all pushes');
    assert(left.jwt === right.jwt, 'jwt must be the same for all pushes with the same clientID');
    assert(left.push.schemaVersion === right.push.schemaVersion, 'schemaVersion must be the same for all pushes with the same clientID');
    assert(left.push.pushVersion === right.push.pushVersion, 'pushVersion must be the same for all pushes with the same clientID');
}
//# sourceMappingURL=pusher.js.map