import { Database } from "../../../../zqlite/src/db.js";
import { StatementRunner } from "../../db/statements.js";
import instruments from "../../observability/view-syncer-instruments.js";
import { PROTOCOL_VERSION, } from "../change-streamer/change-streamer.js";
import { RunningState } from "../running-state.js";
import { ChangeProcessor } from "./change-processor.js";
import { Notifier } from "./notifier.js";
import { getSubscriptionState } from "./schema/replication-state.js";
/**
 * The {@link IncrementalSyncer} manages a logical replication stream from upstream,
 * handling application lifecycle events (start, stop) and retrying the
 * connection with exponential backoff. The actual handling of the logical
 * replication messages is done by the {@link ChangeProcessor}.
 */
export class IncrementalSyncer {
    #taskID;
    #id;
    #changeStreamer;
    #replica;
    #mode;
    #txMode;
    #notifier;
    #state = new RunningState('IncrementalSyncer');
    constructor(taskID, id, changeStreamer, replica, mode) {
        this.#taskID = taskID;
        this.#id = id;
        this.#changeStreamer = changeStreamer;
        this.#replica = new StatementRunner(replica);
        this.#mode = mode;
        this.#txMode = mode === 'serving' ? 'CONCURRENT' : 'IMMEDIATE';
        this.#notifier = new Notifier();
    }
    async run(lc) {
        lc.info?.(`Starting IncrementalSyncer`);
        const { watermark: initialWatermark } = getSubscriptionState(this.#replica);
        // Notify any waiting subscribers that the replica is ready to be read.
        this.#notifier.notifySubscribers();
        while (this.#state.shouldRun()) {
            const { replicaVersion, watermark } = getSubscriptionState(this.#replica);
            const processor = new ChangeProcessor(this.#replica, this.#txMode, (lc, err) => this.stop(lc, err));
            let downstream;
            let unregister = () => { };
            let err;
            try {
                downstream = await this.#changeStreamer.subscribe({
                    protocolVersion: PROTOCOL_VERSION,
                    taskID: this.#taskID,
                    id: this.#id,
                    mode: this.#mode,
                    watermark,
                    replicaVersion,
                    initial: watermark === initialWatermark,
                });
                this.#state.resetBackoff();
                unregister = this.#state.cancelOnStop(downstream);
                for await (const message of downstream) {
                    instruments.counters.replicationEvents.add(1);
                    switch (message[0]) {
                        case 'status':
                            // Used for checking if a replica can be caught up. Not
                            // relevant here.
                            lc.debug?.(`Received initial status`, message[1]);
                            break;
                        case 'error':
                            // Unrecoverable error. Stop the service.
                            this.stop(lc, message[1]);
                            break;
                        default:
                            if (processor.processMessage(lc, message)) {
                                this.#notifier.notifySubscribers({ state: 'version-ready' });
                            }
                    }
                }
                processor.abort(lc);
            }
            catch (e) {
                err = e;
                processor.abort(lc);
            }
            finally {
                downstream?.cancel();
                unregister();
            }
            await this.#state.backoff(lc, err);
        }
        lc.info?.('IncrementalSyncer stopped');
    }
    subscribe() {
        return this.#notifier.subscribe();
    }
    stop(lc, err) {
        this.#state.stop(lc, err);
    }
}
//# sourceMappingURL=incremental-sync.js.map