import { unreachable } from "../../../../shared/src/asserts.js";
import { assertJSONValue, } from "../../../../shared/src/json.js";
import { promiseVoid } from "../../../../shared/src/resolved-promises.js";
import * as v from "../../../../shared/src/valita.js";
import { rowSchema } from "../../../../zero-protocol/src/data.js";
import { primaryKeyValueRecordSchema } from "../../../../zero-protocol/src/primary-key.js";
import { getLogLevel } from "../../types/error-for-client.js";
import { getErrorForClientIfSchemaVersionNotSupported, } from "../../types/schema-versions.js";
import { upstreamSchema } from "../../types/shards.js";
import { cmpVersions, cookieToVersion, versionToCookie, versionToNullableCookie, } from "./schema/types.js";
import instruments from "../../observability/view-syncer-instruments.js";
const NOOP = {
    addPatch: () => promiseVoid,
    cancel: () => promiseVoid,
    end: () => promiseVoid,
};
/** Wraps PokeHandlers for multiple clients in a single PokeHandler. */
export function startPoke(clients, tentativeVersion, schemaVersions) {
    const start = performance.now();
    instruments.counters.pokeTransactions.add(1);
    const pokers = clients.map(c => c.startPoke(tentativeVersion, schemaVersions));
    // Promise.allSettled() ensures that a failed (e.g. disconnected) client
    // does not prevent other clients from receiving the pokes. However, the
    // rate (per client group) will be limited by the slowest connection.
    return {
        addPatch: async (patch) => {
            instruments.counters.rowsPoked.add(1);
            await Promise.allSettled(pokers.map(poker => poker.addPatch(patch)));
        },
        cancel: async () => {
            await Promise.allSettled(pokers.map(poker => poker.cancel()));
        },
        end: async (finalVersion) => {
            const elapsed = performance.now() - start;
            instruments.histograms.pokeTime.record(elapsed);
            await Promise.allSettled(pokers.map(poker => poker.end(finalVersion)));
        },
    };
}
// Semi-arbitrary threshold at which poke body parts are flushed.
// When row size is being computed, that should be used as a threshold instead.
const PART_COUNT_FLUSH_THRESHOLD = 100;
/**
 * Handles a single `ViewSyncer` connection.
 */
export class ClientHandler {
    #clientGroupID;
    clientID;
    wsID;
    #zeroClientsTable;
    #lc;
    #downstream;
    #baseVersion;
    #schemaVersion;
    constructor(lc, clientGroupID, clientID, wsID, shard, baseCookie, schemaVersion, downstream) {
        lc.debug?.('new client handler');
        this.#clientGroupID = clientGroupID;
        this.clientID = clientID;
        this.wsID = wsID;
        this.#zeroClientsTable = `${upstreamSchema(shard)}.clients`;
        this.#lc = lc;
        this.#downstream = downstream;
        this.#baseVersion = cookieToVersion(baseCookie);
        this.#schemaVersion = schemaVersion;
    }
    version() {
        return this.#baseVersion;
    }
    async #push(msg) {
        const { result } = this.#downstream.push(msg);
        await result;
    }
    fail(e) {
        this.#lc[getLogLevel(e)]?.(`view-syncer closing connection with error: ${String(e)}`, e);
        this.#downstream.fail(e instanceof Error ? e : new Error(String(e)));
    }
    close(reason) {
        this.#lc.debug?.(`view-syncer closing connection: ${reason}`);
        this.#downstream.cancel();
    }
    startPoke(tentativeVersion, schemaVersions) {
        const pokeID = versionToCookie(tentativeVersion);
        const lc = this.#lc.withContext('pokeID', pokeID);
        if (schemaVersions && this.#schemaVersion) {
            const schemaVersionError = getErrorForClientIfSchemaVersionNotSupported(this.#schemaVersion, schemaVersions);
            if (schemaVersionError) {
                this.fail(schemaVersionError);
                return NOOP;
            }
        }
        if (cmpVersions(this.#baseVersion, tentativeVersion) >= 0) {
            lc.info?.(`already caught up, not sending poke.`);
            return NOOP;
        }
        const baseCookie = versionToNullableCookie(this.#baseVersion);
        const cookie = versionToCookie(tentativeVersion);
        lc.info?.(`starting poke from ${baseCookie} to ${cookie}`);
        const pokeStart = { pokeID, baseCookie };
        if (schemaVersions) {
            pokeStart.schemaVersions = schemaVersions;
        }
        let pokeStarted = false;
        let body;
        let partCount = 0;
        const ensureBody = async () => {
            if (!pokeStarted) {
                await this.#push(['pokeStart', pokeStart]);
                pokeStarted = true;
            }
            return (body ??= { pokeID });
        };
        const flushBody = async () => {
            if (body) {
                await this.#push(['pokePart', body]);
                body = undefined;
                partCount = 0;
            }
        };
        const addPatch = async (patchToVersion) => {
            const { patch, toVersion } = patchToVersion;
            if (cmpVersions(toVersion, this.#baseVersion) <= 0) {
                return;
            }
            const body = await ensureBody();
            const { type, op } = patch;
            switch (type) {
                case 'query': {
                    const patches = patch.clientID
                        ? ((body.desiredQueriesPatches ??= {})[patch.clientID] ??= [])
                        : (body.gotQueriesPatch ??= []);
                    if (op === 'put') {
                        const { ast } = patch;
                        patches.push({ op, hash: patch.id, ast });
                    }
                    else {
                        patches.push({ op, hash: patch.id });
                    }
                    break;
                }
                case 'row':
                    if (patch.id.table === this.#zeroClientsTable) {
                        this.#updateLMIDs((body.lastMutationIDChanges ??= {}), patch);
                    }
                    else {
                        (body.rowsPatch ??= []).push(makeRowPatch(patch));
                    }
                    break;
                default:
                    unreachable(patch);
            }
            if (++partCount >= PART_COUNT_FLUSH_THRESHOLD) {
                await flushBody();
            }
        };
        return {
            addPatch: async (patchToVersion) => {
                try {
                    await addPatch(patchToVersion);
                }
                catch (e) {
                    this.#downstream.fail(e instanceof Error ? e : new Error(String(e)));
                }
            },
            cancel: async () => {
                if (pokeStarted) {
                    await this.#push(['pokeEnd', { pokeID, cookie: '', cancel: true }]);
                }
            },
            end: async (finalVersion) => {
                const cookie = versionToCookie(finalVersion);
                if (!pokeStarted) {
                    if (cmpVersions(this.#baseVersion, finalVersion) === 0) {
                        return; // Nothing changed and nothing was sent.
                    }
                    await this.#push(['pokeStart', pokeStart]);
                }
                else if (cmpVersions(this.#baseVersion, finalVersion) >= 0) {
                    // Sanity check: If the poke was started, the finalVersion
                    // must be > #baseVersion.
                    throw new Error(`Patches were sent but finalVersion ${finalVersion} is ` +
                        `not greater than baseVersion ${this.#baseVersion}`);
                }
                await flushBody();
                await this.#push(['pokeEnd', { pokeID, cookie }]);
                this.#baseVersion = finalVersion;
            },
        };
    }
    async sendDeleteClients(lc, deletedClientIDs, deletedClientGroupIDs) {
        const deleteClientsBody = {};
        if (deletedClientIDs.length > 0) {
            deleteClientsBody.clientIDs = deletedClientIDs;
        }
        if (deletedClientGroupIDs.length > 0) {
            deleteClientsBody.clientGroupIDs = deletedClientGroupIDs;
        }
        lc.debug?.('sending deleteClients', deleteClientsBody);
        await this.#push(['deleteClients', deleteClientsBody]);
    }
    sendInspectResponse(lc, response) {
        lc.debug?.('sending inspect response', response);
        this.#downstream.push(['inspect', response]);
    }
    #updateLMIDs(lmids, patch) {
        if (patch.op === 'put') {
            const row = ensureSafeJSON(patch.contents);
            const { clientGroupID, clientID, lastMutationID } = v.parse(row, lmidRowSchema, 'passthrough');
            if (clientGroupID !== this.#clientGroupID) {
                this.#lc.error?.(`Received clients row for wrong clientGroupID. Ignoring.`, clientGroupID);
            }
            else {
                lmids[clientID] = lastMutationID;
            }
        }
        else {
            // The 'constrain' and 'del' ops for clients can be ignored.
            patch.op;
        }
    }
}
// Note: The {APP_ID}_{SHARD_ID}.clients table is set up in replicator/initial-sync.ts.
const lmidRowSchema = v.object({
    clientGroupID: v.string(),
    clientID: v.string(),
    lastMutationID: v.number(), // Actually returned as a bigint, but converted by ensureSafeJSON().
});
function makeRowPatch(patch) {
    const { op, id: { table: tableName, rowKey: id }, } = patch;
    switch (op) {
        case 'put':
            return {
                op: 'put',
                tableName,
                value: v.parse(ensureSafeJSON(patch.contents), rowSchema),
            };
        case 'del':
            return {
                op,
                tableName,
                id: v.parse(id, primaryKeyValueRecordSchema),
            };
        default:
            unreachable(op);
    }
}
/**
 * Column values of type INT8 are returned as the `bigint` from the
 * Postgres library. These are converted to `number` if they are within
 * the safe Number range, allowing the protocol to support numbers larger
 * than 32-bits. Values outside of the safe number range (e.g. > 2^53) will
 * result in an Error.
 */
export function ensureSafeJSON(row) {
    const modified = Object.entries(row)
        .filter(([k, v]) => {
        if (typeof v === 'bigint') {
            if (v >= Number.MIN_SAFE_INTEGER && v <= Number.MAX_SAFE_INTEGER) {
                return true; // send this entry onto the next map() step.
            }
            throw new Error(`Value of "${k}" exceeds safe Number range (${v})`);
        }
        else if (typeof v === 'object') {
            assertJSONValue(v);
        }
        return false;
    })
        .map(([k, v]) => [k, Number(v)]);
    return modified.length
        ? { ...row, ...Object.fromEntries(modified) }
        : row;
}
//# sourceMappingURL=client-handler.js.map