import type { LogContext } from '@rocicorp/logger';
import type { JWTPayload } from 'jose';
import type { ChangeDesiredQueriesMessage } from '../../../../zero-protocol/src/change-desired-queries.ts';
import type { CloseConnectionMessage } from '../../../../zero-protocol/src/close-connection.ts';
import type { InitConnectionMessage } from '../../../../zero-protocol/src/connect.ts';
import type { DeleteClientsMessage } from '../../../../zero-protocol/src/delete-clients.ts';
import type { Downstream } from '../../../../zero-protocol/src/down.ts';
import type { InspectUpMessage } from '../../../../zero-protocol/src/inspect-up.ts';
import type { PostgresDB } from '../../types/pg.ts';
import type { ShardID } from '../../types/shards.ts';
import type { Source } from '../../types/streams.ts';
import { Subscription } from '../../types/subscription.ts';
import type { ReplicaState } from '../replicator/replicator.ts';
import type { ActivityBasedService } from '../service.ts';
import type { DrainCoordinator } from './drain-coordinator.ts';
import { PipelineDriver } from './pipeline-driver.ts';
export type TokenData = {
    readonly raw: string;
    readonly decoded: JWTPayload;
};
export type SyncContext = {
    readonly clientID: string;
    readonly wsID: string;
    readonly baseCookie: string | null;
    readonly protocolVersion: number;
    readonly schemaVersion: number | null;
    readonly tokenData: TokenData | undefined;
};
export interface ViewSyncer {
    initConnection(ctx: SyncContext, msg: InitConnectionMessage): Source<Downstream>;
    changeDesiredQueries(ctx: SyncContext, msg: ChangeDesiredQueriesMessage): Promise<void>;
    deleteClients(ctx: SyncContext, msg: DeleteClientsMessage): Promise<void>;
    closeConnection(ctx: SyncContext, msg: CloseConnectionMessage): Promise<void>;
    inspect(context: SyncContext, msg: InspectUpMessage): Promise<void>;
}
type SetTimeout = (fn: (...args: unknown[]) => void, delay?: number) => ReturnType<typeof setTimeout>;
export declare class ViewSyncerService implements ViewSyncer, ActivityBasedService {
    #private;
    readonly id: string;
    /**
     * The {@linkcode maxRowCount} is used for the eviction of inactive queries.
     * An inactive query is a query that is no longer desired but is kept alive
     * due to its TTL. When the number of rows in the CVR exceeds
     * {@linkcode maxRowCount} we keep removing inactive queries (even if they are
     * not expired yet) until the actual row count is below the max row count.
     *
     * There is no guarantee that the number of rows in the CVR will be below this
     * if there are active queries that have a lot of rows.
     */
    maxRowCount: number;
    constructor(lc: LogContext, shard: ShardID, taskID: string, clientGroupID: string, db: PostgresDB, pipelineDriver: PipelineDriver, versionChanges: Subscription<ReplicaState>, drainCoordinator: DrainCoordinator, slowHydrateThreshold: number, keepaliveMs?: number, maxRowCount?: number, setTimeoutFn?: SetTimeout);
    run(): Promise<void>;
    /**
     * Guarantees that the ViewSyncer will remain running for at least
     * its configured `keepaliveMs`. This is called when establishing a
     * new connection to ensure that its associated ViewSyncer isn't
     * shutdown before it receives the connection.
     *
     * @return `true` if the ViewSyncer will stay alive, `false` if the
     *         ViewSyncer is shutting down.
     */
    keepalive(): boolean;
    initConnection(ctx: SyncContext, initConnectionMessage: InitConnectionMessage): Source<Downstream>;
    changeDesiredQueries(ctx: SyncContext, msg: ChangeDesiredQueriesMessage): Promise<void>;
    deleteClients(ctx: SyncContext, msg: DeleteClientsMessage): Promise<void>;
    closeConnection(ctx: SyncContext, msg: CloseConnectionMessage): Promise<void>;
    inspect(context: SyncContext, msg: InspectUpMessage): Promise<void>;
    stop(): Promise<void>;
}
export declare function pickToken(lc: LogContext, previousToken: JWTPayload | undefined, newToken: JWTPayload | undefined): JWTPayload | undefined;
export declare class Timer {
    #private;
    start(): this;
    startLap(): void;
    elapsedLap(): number;
    stopLap(): void;
    /** @returns the total elapsed time */
    stop(): number;
    /**
     * @returns the elapsed time. This can be called while the Timer is running
     *          or after it has been stopped.
     */
    totalElapsed(): number;
}
export {};
//# sourceMappingURL=view-syncer.d.ts.map