import { LogContext } from '@rocicorp/logger';
import postgres, { type PostgresType } from 'postgres';
import type { ValueType } from '../../../zero-protocol/src/client-schema.ts';
import { type JSONValue } from './bigint-json.ts';
export declare function timestampToFpMillis(timestamp: string): number;
declare function serializeTimestamp(val: unknown): string;
declare function dateToUTCMidnight(date: string): number;
/**
 * The (javascript) types of objects that can be returned by our configured
 * Postgres clients. For initial-sync, these comes from the postgres.js client:
 *
 * https://github.com/porsager/postgres/blob/master/src/types.js
 *
 * and for the replication stream these come from the the node-postgres client:
 *
 * https://github.com/brianc/node-pg-types/blob/master/lib/textParsers.js
 */
export type PostgresValueType = JSONValue | Uint8Array;
/**
 * Configures types for the Postgres.js client library (`postgres`).
 *
 * @param jsonAsString Keep JSON / JSONB values as strings instead of parsing.
 */
export declare const postgresTypeConfig: (jsonAsString?: "json-as-string" | undefined) => {
    types: {
        bigint: postgres.PostgresType<bigint>;
        json: {
            to: number;
            from: number[];
            serialize: typeof import("./bigint-json.ts").stringify;
            parse: typeof import("./bigint-json.ts").parse;
        };
        timestamp: {
            to: number;
            from: number[];
            serialize: typeof serializeTimestamp;
            parse: typeof timestampToFpMillis;
        };
        date: {
            to: number;
            from: number[];
            serialize: (x: string | Date) => string;
            parse: typeof dateToUTCMidnight;
        };
        numeric: {
            to: number;
            from: number[];
            serialize: (x: number) => string;
            parse: (x: string | number) => number;
        };
    };
};
export type PostgresDB = postgres.Sql<{
    bigint: bigint;
    json: JSONValue;
}>;
export type PostgresTransaction = postgres.TransactionSql<{
    bigint: bigint;
    json: JSONValue;
}>;
export declare function pgClient(lc: LogContext, connectionURI: string, options?: postgres.Options<{
    bigint: PostgresType<bigint>;
    json: PostgresType<JSONValue>;
}>, jsonAsString?: 'json-as-string'): PostgresDB;
export declare const typeNameByOID: Record<number, string>;
export declare const pgToZqlNumericTypeMap: Readonly<{
    smallint: "number";
    integer: "number";
    int: "number";
    int2: "number";
    int4: "number";
    int8: "number";
    bigint: "number";
    smallserial: "number";
    serial: "number";
    serial2: "number";
    serial4: "number";
    serial8: "number";
    bigserial: "number";
    decimal: "number";
    numeric: "number";
    real: "number";
    'double precision': "number";
    float: "number";
    float4: "number";
    float8: "number";
}>;
export declare function isPgNumberType(pgType: string): boolean;
export declare const pgToZqlStringTypeMap: Readonly<{
    bpchar: "string";
    character: "string";
    'character varying': "string";
    text: "string";
    uuid: "string";
    varchar: "string";
}>;
export declare function isPgStringType(pgType: string): boolean;
export declare const pgToZqlTypeMap: Readonly<{
    bool: "boolean";
    boolean: "boolean";
    json: "json";
    jsonb: "json";
    bpchar: "string";
    character: "string";
    'character varying': "string";
    text: "string";
    uuid: "string";
    varchar: "string";
    date: "number";
    timestamp: "number";
    timestamptz: "number";
    'timestamp with time zone': "number";
    'timestamp without time zone': "number";
    smallint: "number";
    integer: "number";
    int: "number";
    int2: "number";
    int4: "number";
    int8: "number";
    bigint: "number";
    smallserial: "number";
    serial: "number";
    serial2: "number";
    serial4: "number";
    serial8: "number";
    bigserial: "number";
    decimal: "number";
    numeric: "number";
    real: "number";
    'double precision': "number";
    float: "number";
    float4: "number";
    float8: "number";
}>;
export declare function dataTypeToZqlValueType(pgType: string, isEnum: boolean): ValueType | undefined;
export {};
//# sourceMappingURL=pg.d.ts.map