import { PreciseDate } from '@google-cloud/precise-date';
import { OID } from '@postgresql-typed/oids';
import { LogContext } from '@rocicorp/logger';
import postgres, {} from 'postgres';
import { randInt } from "../../../shared/src/rand.js";
import { BigIntJSON } from "./bigint-json.js";
import { DATE, JSON, JSONB, NUMERIC, TIMESTAMP, TIMESTAMPTZ, } from "./pg-types.js";
// exported for testing.
export function timestampToFpMillis(timestamp) {
    // Convert from PG's time string, e.g. "1999-01-08 12:05:06+00" to "Z"
    // format expected by PreciseDate.
    timestamp = timestamp.replace(' ', 'T');
    const positiveOffset = timestamp.includes('+');
    const tzSplitIndex = positiveOffset
        ? timestamp.lastIndexOf('+')
        : timestamp.indexOf('-', timestamp.indexOf('T'));
    const timezoneOffset = tzSplitIndex === -1 ? undefined : timestamp.substring(tzSplitIndex);
    const tsWithoutTimezone = (tzSplitIndex === -1 ? timestamp : timestamp.substring(0, tzSplitIndex)) +
        'Z';
    try {
        // PreciseDate does not return microsecond precision unless the provided
        // timestamp is in UTC time so we need to add the timezone offset back in.
        const fullTime = new PreciseDate(tsWithoutTimezone).getFullTime();
        const millis = Number(fullTime / 1000000n);
        const nanos = Number(fullTime % 1000000n);
        const ret = millis + nanos * 1e-6; // floating point milliseconds
        // add back in the timezone offset
        if (timezoneOffset) {
            const [hours, minutes] = timezoneOffset.split(':');
            const offset = Math.abs(Number(hours)) * 60 + (minutes ? Number(minutes) : 0);
            const offsetMillis = offset * 60 * 1_000;
            // If it is a positive offset, we subtract the offset from the UTC
            // because we passed in the "local time" as if it was UTC.
            // The opposite is true for negative offsets.
            return positiveOffset ? ret - offsetMillis : ret + offsetMillis;
        }
        return ret;
    }
    catch (e) {
        throw new Error(`Error parsing ${timestamp}`, { cause: e });
    }
}
function serializeTimestamp(val) {
    switch (typeof val) {
        case 'string':
            return val; // Let Postgres parse it
        case 'number': {
            if (Number.isInteger(val)) {
                return new PreciseDate(val).toISOString();
            }
            // Convert floating point to bigint nanoseconds.
            const nanoseconds = 1000000n * BigInt(Math.trunc(val)) +
                BigInt(Math.trunc((val % 1) * 1e6));
            return new PreciseDate(nanoseconds).toISOString();
        }
        // Note: Don't support bigint inputs until we decide what the semantics are (e.g. micros vs nanos)
        // case 'bigint':
        //   return new PreciseDate(val).toISOString();
        default:
            if (val instanceof Date) {
                return val.toISOString();
            }
    }
    throw new Error(`Unsupported type "${typeof val}" for timestamp: ${val}`);
}
function dateToUTCMidnight(date) {
    const d = new Date(date);
    return Date.UTC(d.getUTCFullYear(), d.getUTCMonth(), d.getUTCDate());
}
/**
 * Configures types for the Postgres.js client library (`postgres`).
 *
 * @param jsonAsString Keep JSON / JSONB values as strings instead of parsing.
 */
export const postgresTypeConfig = (jsonAsString) => ({
    // Type the type IDs as `number` so that Typescript doesn't complain about
    // referencing external types during type inference.
    types: {
        bigint: postgres.BigInt,
        json: {
            to: JSON,
            from: [JSON, JSONB],
            serialize: BigIntJSON.stringify,
            parse: jsonAsString ? (x) => x : BigIntJSON.parse,
        },
        // Timestamps are converted to PreciseDate objects.
        timestamp: {
            to: TIMESTAMP,
            from: [TIMESTAMP, TIMESTAMPTZ],
            serialize: serializeTimestamp,
            parse: timestampToFpMillis,
        },
        // The DATE type is stored directly as the PG normalized date string.
        date: {
            to: DATE,
            from: [DATE],
            serialize: (x) => (x instanceof Date ? x : new Date(x)).toISOString(),
            parse: dateToUTCMidnight,
        },
        // Returns a `js` number which can lose precision for large numbers.
        // JS number is 53 bits so this should generally not occur.
        // An API will be provided for users to override this type.
        numeric: {
            to: NUMERIC,
            from: [NUMERIC],
            serialize: (x) => String(x), // pg expects a string
            parse: (x) => Number(x),
        },
    },
});
export function pgClient(lc, connectionURI, options, jsonAsString) {
    const onnotice = (n) => {
        // https://www.postgresql.org/docs/current/plpgsql-errors-and-messages.html#PLPGSQL-STATEMENTS-RAISE
        switch (n.severity) {
            case 'NOTICE':
                return; // silenced
            case 'DEBUG':
                lc.debug?.(n);
                return;
            case 'WARNING':
            case 'EXCEPTION':
                lc.error?.(n);
                return;
            case 'LOG':
            case 'INFO':
            default:
                lc.info?.(n);
        }
    };
    const url = new URL(connectionURI);
    const ssl = url.searchParams.get('ssl') ?? url.searchParams.get('sslmode') ?? 'prefer';
    // Set connections to expire between 5 and 10 minutes to free up state on PG.
    const maxLifetimeSeconds = randInt(5 * 60, 10 * 60);
    return postgres(connectionURI, {
        ...postgresTypeConfig(jsonAsString),
        onnotice,
        ['max_lifetime']: maxLifetimeSeconds,
        ssl: ssl === 'disable' || ssl === 'false' ? false : ssl,
        ...options,
    });
}
export const typeNameByOID = Object.fromEntries(Object.entries(OID).map(([name, oid]) => [
    oid,
    name.startsWith('_') ? `${name.substring(1)}[]` : name,
]));
Object.freeze(typeNameByOID);
export const pgToZqlNumericTypeMap = Object.freeze({
    'smallint': 'number',
    'integer': 'number',
    'int': 'number',
    'int2': 'number',
    'int4': 'number',
    'int8': 'number',
    'bigint': 'number',
    'smallserial': 'number',
    'serial': 'number',
    'serial2': 'number',
    'serial4': 'number',
    'serial8': 'number',
    'bigserial': 'number',
    'decimal': 'number',
    'numeric': 'number',
    'real': 'number',
    'double precision': 'number',
    'float': 'number',
    'float4': 'number',
    'float8': 'number',
});
export function isPgNumberType(pgType) {
    return Object.hasOwn(pgToZqlNumericTypeMap, formatTypeForLookup(pgType));
}
export const pgToZqlStringTypeMap = Object.freeze({
    'bpchar': 'string',
    'character': 'string',
    'character varying': 'string',
    'text': 'string',
    'uuid': 'string',
    'varchar': 'string',
});
export function isPgStringType(pgType) {
    return Object.hasOwn(pgToZqlStringTypeMap, formatTypeForLookup(pgType));
}
export const pgToZqlTypeMap = Object.freeze({
    // Numeric types
    ...pgToZqlNumericTypeMap,
    // Date/Time types
    'date': 'number',
    'timestamp': 'number',
    'timestamptz': 'number',
    'timestamp with time zone': 'number',
    'timestamp without time zone': 'number',
    // String types
    ...pgToZqlStringTypeMap,
    // Boolean types
    'bool': 'boolean',
    'boolean': 'boolean',
    'json': 'json',
    'jsonb': 'json',
    // TODO: Add support for these.
    // 'bytea':
});
export function dataTypeToZqlValueType(pgType, isEnum) {
    const valueType = pgToZqlTypeMap[formatTypeForLookup(pgType)];
    if (valueType === undefined && isEnum) {
        return 'string';
    }
    return valueType;
}
// Strips args (i.e. (32) in char(32)) and lowercases.
function formatTypeForLookup(pgType) {
    const startOfArgs = pgType.indexOf('(');
    if (startOfArgs === -1) {
        return pgType.toLocaleLowerCase();
    }
    return pgType.toLocaleLowerCase().substring(0, startOfArgs);
}
//# sourceMappingURL=pg.js.map