import { LogContext } from '@rocicorp/logger';
import { resolver } from '@rocicorp/resolver';
import {} from 'jose';
import { pid } from 'node:process';
import { MessagePort } from 'node:worker_threads';
import { WebSocketServer } from 'ws';
import { promiseVoid } from "../../../shared/src/resolved-promises.js";
import { ErrorKind } from "../../../zero-protocol/src/error-kind.js";
import { verifyToken } from "../auth/jwt.js";
import {} from "../config/zero-config.js";
import { ServiceRunner } from "../services/runner.js";
import { DrainCoordinator } from "../services/view-syncer/drain-coordinator.js";
import { Subscription } from "../types/subscription.js";
import { installWebSocketReceiver } from "../types/websocket-handoff.js";
import { Connection, sendError } from "./connection.js";
import { createNotifierFrom, subscribeTo } from "./replicator.js";
import { SyncerWsMessageHandler } from "./syncer-ws-message-handler.js";
/**
 * The Syncer worker receives websocket handoffs for "/sync" connections
 * from the Dispatcher in the main thread, and creates websocket
 * {@link Connection}s with a corresponding {@link ViewSyncer}, {@link Mutagen},
 * and {@link Subscription} to version notifications from the Replicator
 * worker.
 */
export class Syncer {
    id = `syncer-${pid}`;
    #lc;
    #viewSyncers;
    #mutagens;
    #pushers;
    #connections = new Map();
    #drainCoordinator = new DrainCoordinator();
    #parent;
    #wss;
    #stopped = resolver();
    #authConfig;
    constructor(lc, config, viewSyncerFactory, mutagenFactory, pusherFactory, parent) {
        this.#authConfig = config.auth;
        // Relays notifications from the parent thread subscription
        // to ViewSyncers within this thread.
        const notifier = createNotifierFrom(lc, parent);
        subscribeTo(lc, parent);
        this.#lc = lc;
        this.#viewSyncers = new ServiceRunner(lc, id => viewSyncerFactory(id, notifier.subscribe(), this.#drainCoordinator), v => v.keepalive());
        this.#mutagens = new ServiceRunner(lc, mutagenFactory);
        if (pusherFactory) {
            this.#pushers = new ServiceRunner(lc, pusherFactory);
        }
        this.#parent = parent;
        this.#wss = new WebSocketServer({ noServer: true });
        installWebSocketReceiver(lc, this.#wss, this.#createConnection, this.#parent);
    }
    #createConnection = async (ws, params) => {
        this.#lc.debug?.('creating connection', params.clientGroupID, params.clientID);
        const { clientID, clientGroupID, auth, userID } = params;
        const existing = this.#connections.get(clientID);
        if (existing) {
            this.#lc.debug?.(`client ${clientID} already connected, closing existing connection`);
            existing.close(`replaced by ${params.wsID}`);
        }
        let decodedToken;
        if (auth) {
            try {
                decodedToken = await verifyToken(this.#authConfig, auth, {
                    subject: userID,
                });
                this.#lc.debug?.(`Received auth token ${auth} for clientID ${clientID}, decoded: ${JSON.stringify(decodedToken)}`);
            }
            catch (e) {
                sendError(this.#lc, ws, {
                    kind: ErrorKind.AuthInvalidated,
                    message: `Failed to decode auth token: ${String(e)}`,
                }, e);
                ws.close(3000, 'Failed to decode JWT');
                return;
            }
        }
        else {
            this.#lc.debug?.(`No auth token received for clientID ${clientID}`);
        }
        const connection = new Connection(this.#lc, params, ws, new SyncerWsMessageHandler(this.#lc, params, auth !== undefined && decodedToken !== undefined
            ? {
                raw: auth,
                decoded: decodedToken,
            }
            : undefined, this.#viewSyncers.getService(clientGroupID), this.#mutagens.getService(clientGroupID), this.#pushers?.getService(clientGroupID)), () => {
            if (this.#connections.get(clientID) === connection) {
                this.#connections.delete(clientID);
            }
        });
        this.#connections.set(clientID, connection);
        connection.init();
        if (params.initConnectionMsg) {
            this.#lc.debug?.('handling init connection message from sec header', params.clientGroupID, params.clientID);
            await connection.handleInitConnection(JSON.stringify(params.initConnectionMsg));
        }
    };
    run() {
        return this.#stopped.promise;
    }
    /**
     * Graceful shutdown involves shutting down view syncers one at a time, pausing
     * for the duration of view syncer's hydration between each one. This paces the
     * disconnects to avoid creating a backlog of hydrations in the receiving server
     * when the clients reconnect.
     */
    async drain() {
        const start = Date.now();
        this.#lc.info?.(`draining ${this.#viewSyncers.size} view-syncers`);
        this.#drainCoordinator.drainNextIn(0);
        while (this.#viewSyncers.size) {
            await this.#drainCoordinator.forceDrainTimeout;
            // Pick an arbitrary view syncer to force drain.
            for (const vs of this.#viewSyncers.getServices()) {
                this.#lc.debug?.(`draining view-syncer ${vs.id} (forced)`);
                // When this drain or an elective drain completes, the forceDrainTimeout will
                // resolve after the next drain interval.
                void vs.stop();
                break;
            }
        }
        this.#lc.info?.(`finished draining (${Date.now() - start} ms)`);
    }
    stop() {
        this.#wss.close();
        this.#stopped.resolve();
        return promiseVoid;
    }
}
//# sourceMappingURL=syncer.js.map