import { type LogLevel } from '@rocicorp/logger';
import type { ReadonlyJSONValue } from '../../shared/src/json.ts';
import * as v from '../../shared/src/valita.ts';
import { pushParamsSchema, type PushResponse } from '../../zero-protocol/src/push.ts';
import type { CustomMutatorDefs } from './custom.ts';
export type Params = v.Infer<typeof pushParamsSchema>;
export interface TransactionProviderHooks {
    updateClientMutationID: () => Promise<{
        lastMutationID: number | bigint;
    }>;
}
export interface TransactionProviderInput {
    upstreamSchema: string;
    clientGroupID: string;
    clientID: string;
    mutationID: number;
}
/**
 * Defines the abstract interface for a database that PushProcessor can execute
 * transactions against.
 */
export interface Database<T> {
    transaction: <R>(callback: (tx: T, transactionHooks: TransactionProviderHooks) => Promise<R>, transactionInput: TransactionProviderInput) => Promise<R>;
}
type ExtractTransactionType<D> = D extends Database<infer T> ? T : never;
export declare class PushProcessor<D extends Database<ExtractTransactionType<D>>, MD extends CustomMutatorDefs<ExtractTransactionType<D>>> {
    #private;
    constructor(dbProvider: D, logLevel?: LogLevel);
    /**
     * Processes a push request from zero-cache.
     * This function will parse the request, check the protocol version, and process each mutation in the request.
     * - If a mutation is out of order: processing will stop and an error will be returned. The zero client will retry the mutation.
     * - If a mutation has already been processed: it will be skipped and the processing will continue.
     * - If a mutation receives an application error: it will be skipped, the error will be returned to the client, and processing will continue.
     *
     * @param mutators the custom mutators for the application
     * @param queryString the query string from the request sent by zero-cache. This will include zero's postgres schema name and appID.
     * @param body the body of the request sent by zero-cache as a JSON object.
     */
    process(mutators: MD, queryString: URLSearchParams | Record<string, string>, body: ReadonlyJSONValue): Promise<PushResponse>;
    /**
     * This override gets the query string and the body from a Request object.
     *
     * @param mutators the custom mutators for the application
     * @param request A `Request` object.
     */
    process(mutators: MD, request: Request): Promise<PushResponse>;
}
export {};
//# sourceMappingURL=push-processor.d.ts.map