import { LogContext } from '@rocicorp/logger';
import { assert } from "../../shared/src/asserts.js";
import * as v from "../../shared/src/valita.js";
import { MutationAlreadyProcessedError } from "../../zero-cache/src/services/mutagen/mutagen.js";
import { pushBodySchema, pushParamsSchema, } from "../../zero-protocol/src/push.js";
import { splitMutatorKey } from "../../zql/src/mutate/custom.js";
import { createLogContext } from "./logging.js";
export class PushProcessor {
    #dbProvider;
    #lc;
    constructor(dbProvider, logLevel = 'info') {
        this.#dbProvider = dbProvider;
        this.#lc = createLogContext(logLevel).withContext('PushProcessor');
    }
    async process(mutators, queryOrQueryString, body) {
        let queryString;
        if (queryOrQueryString instanceof Request) {
            const url = new URL(queryOrQueryString.url);
            queryString = url.searchParams;
            body = await queryOrQueryString.json();
        }
        else {
            queryString = queryOrQueryString;
        }
        const req = v.parse(body, pushBodySchema);
        if (queryString instanceof URLSearchParams) {
            queryString = Object.fromEntries(queryString);
        }
        const queryParams = v.parse(queryString, pushParamsSchema, 'passthrough');
        if (req.pushVersion !== 1) {
            this.#lc.error?.(`Unsupported push version ${req.pushVersion} for clientGroupID ${req.clientGroupID}`);
            return {
                error: 'unsupportedPushVersion',
            };
        }
        const responses = [];
        for (const m of req.mutations) {
            const res = await this.#processMutation(mutators, queryParams, req, m);
            responses.push(res);
            if ('error' in res.result) {
                break;
            }
        }
        return {
            mutations: responses,
        };
    }
    async #processMutation(mutators, params, req, m) {
        try {
            return await this.#processMutationImpl(mutators, params, req, m, false);
        }
        catch (e) {
            if (e instanceof OutOfOrderMutation) {
                this.#lc.error?.(e);
                return {
                    id: {
                        clientID: m.clientID,
                        id: m.id,
                    },
                    result: {
                        error: 'oooMutation',
                        details: e.message,
                    },
                };
            }
            if (e instanceof MutationAlreadyProcessedError) {
                this.#lc.warn?.(e);
                return {
                    id: {
                        clientID: m.clientID,
                        id: m.id,
                    },
                    result: {
                        error: 'alreadyProcessed',
                        details: e.message,
                    },
                };
            }
            const ret = await this.#processMutationImpl(mutators, params, req, m, true);
            if ('error' in ret.result) {
                this.#lc.error?.(`Error ${ret.result.error} processing mutation ${m.id} for client ${m.clientID}: ${ret.result.details}`, e);
                return ret;
            }
            this.#lc.error?.(`Unexpected error processing mutation ${m.id} for client ${m.clientID}`, e);
            return {
                id: ret.id,
                result: {
                    error: 'app',
                    details: e instanceof Error
                        ? e.message
                        : 'exception was not of type `Error`',
                },
            };
        }
    }
    #processMutationImpl(mutators, params, req, m, errorMode) {
        if (m.type === 'crud') {
            throw new Error('crud mutators are deprecated in favor of custom mutators.');
        }
        return this.#dbProvider.transaction(async (dbTx, transactionHooks) => {
            await this.#checkAndIncrementLastMutationID(this.#lc, transactionHooks, m.clientID, m.id);
            if (!errorMode) {
                await this.#dispatchMutation(dbTx, mutators, m);
            }
            return {
                id: {
                    clientID: m.clientID,
                    id: m.id,
                },
                result: {},
            };
        }, {
            upstreamSchema: params.schema,
            clientGroupID: req.clientGroupID,
            clientID: m.clientID,
            mutationID: m.id,
        });
    }
    #dispatchMutation(dbTx, mutators, m) {
        const [namespace, name] = splitMutatorKey(m.name);
        if (name === undefined) {
            const mutator = mutators[namespace];
            assert(typeof mutator === 'function', () => `could not find mutator ${m.name}`);
            return mutator(dbTx, m.args[0]);
        }
        const mutatorGroup = mutators[namespace];
        assert(typeof mutatorGroup === 'object', () => `could not find mutators for namespace ${namespace}`);
        const mutator = mutatorGroup[name];
        assert(typeof mutator === 'function', () => `could not find mutator ${m.name}`);
        return mutator(dbTx, m.args[0]);
    }
    async #checkAndIncrementLastMutationID(lc, transactionHooks, clientID, receivedMutationID) {
        lc.debug?.(`Incrementing LMID. Received: ${receivedMutationID}`);
        const { lastMutationID } = await transactionHooks.updateClientMutationID();
        if (receivedMutationID < lastMutationID) {
            throw new MutationAlreadyProcessedError(clientID, receivedMutationID, lastMutationID);
        }
        else if (receivedMutationID > lastMutationID) {
            throw new OutOfOrderMutation(clientID, receivedMutationID, lastMutationID);
        }
        lc.debug?.(`Incremented LMID. Received: ${receivedMutationID}. New: ${lastMutationID}`);
    }
}
class OutOfOrderMutation extends Error {
    constructor(clientID, receivedMutationID, lastMutationID) {
        super(`Client ${clientID} sent mutation ID ${receivedMutationID} but expected ${lastMutationID}`);
    }
}
//# sourceMappingURL=push-processor.js.map