/* eslint-disable @typescript-eslint/no-explicit-any */
import { h64 } from "../../../shared/src/hash.js";
import { mapEntries } from "../../../shared/src/objects.js";
import { normalizeClientSchema, } from "../../../zero-protocol/src/client-schema.js";
import {} from "./table-builder.js";
/**
 * Note: the keys of the `tables` and `relationships` parameters do not matter.
 * You can assign them to any value you like. E.g.,
 *
 * ```ts
 * createSchema({rsdfgafg: table('users')...}, {sdfd: relationships(users, ...)})
 * ```
 */
export function createSchema(options) {
    const retTables = {};
    const retRelationships = {};
    const serverNames = new Set();
    options.tables.forEach(table => {
        const { serverName = table.schema.name } = table.schema;
        if (serverNames.has(serverName)) {
            throw new Error(`Multiple tables reference the name "${serverName}"`);
        }
        serverNames.add(serverName);
        if (retTables[table.schema.name]) {
            throw new Error(`Table "${table.schema.name}" is defined more than once in the schema`);
        }
        retTables[table.schema.name] = table.build();
    });
    options.relationships?.forEach(relationships => {
        if (retRelationships[relationships.name]) {
            throw new Error(`Relationships for table "${relationships.name}" are defined more than once in the schema`);
        }
        retRelationships[relationships.name] = relationships.relationships;
        checkRelationship(relationships.relationships, relationships.name, retTables);
    });
    return {
        tables: retTables,
        relationships: retRelationships,
    };
}
function checkRelationship(relationships, tableName, tables) {
    // TS should be able to check this for us but something is preventing it from happening.
    Object.entries(relationships).forEach(([name, rel]) => {
        let source = tables[tableName];
        rel.forEach(connection => {
            if (!tables[connection.destSchema]) {
                throw new Error(`For relationship "${tableName}"."${name}", destination table "${connection.destSchema}" is missing in the schema`);
            }
            if (!source.columns[connection.sourceField[0]]) {
                throw new Error(`For relationship "${tableName}"."${name}", the source field "${connection.sourceField[0]}" is missing in the table schema "${source.name}"`);
            }
            source = tables[connection.destSchema];
        });
    });
}
export function clientSchemaFrom(schema) {
    const client = {
        tables: mapEntries(schema.tables, (name, { serverName, columns }) => [
            serverName ?? name,
            {
                columns: mapEntries(columns, (name, { serverName, type }) => [
                    serverName ?? name,
                    { type },
                ]),
            },
        ]),
    };
    const clientSchema = normalizeClientSchema(client);
    const hash = h64(JSON.stringify(clientSchema)).toString(36);
    return { clientSchema, hash };
}
//# sourceMappingURL=schema-builder.js.map