export function clientToServer(tables) {
    return createMapperFrom('client', tables);
}
export function serverToClient(tables) {
    return createMapperFrom('server', tables);
}
function createMapperFrom(src, tables) {
    const mapping = new Map(Object.entries(tables).map(([tableName, { serverName: serverTableName, columns }]) => {
        let allColumnsSame = true;
        const names = {};
        for (const [name, { serverName }] of Object.entries(columns)) {
            if (serverName && serverName !== name) {
                allColumnsSame = false;
            }
            if (src === 'client') {
                names[name] = serverName ?? name;
            }
            else {
                names[serverName ?? name] = name;
            }
        }
        return [
            src === 'client' ? tableName : (serverTableName ?? tableName),
            {
                tableName: src === 'client' ? (serverTableName ?? tableName) : tableName,
                columns: names,
                allColumnsSame,
            },
        ];
    }));
    return new NameMapper(mapping);
}
/**
 * Returns an "identity" NameMapper that simply serves the purpose
 * of validating that all table and column names conform to the
 * specified `tablesToColumns` map.
 */
export function validator(tablesToColumns) {
    const identity = new Map([...tablesToColumns.entries()].map(([tableName, columns]) => [
        tableName,
        {
            tableName,
            columns: Object.fromEntries(columns.map(c => [c, c])),
            allColumnsSame: true,
        },
    ]));
    return new NameMapper(identity);
}
export class NameMapper {
    #tables = new Map();
    constructor(tables) {
        this.#tables = tables;
    }
    #getTable(src, ctx) {
        const table = this.#tables.get(src);
        if (!table) {
            throw new Error(`unknown table "${src}" ${!ctx ? '' : `in ${JSON.stringify(ctx)}`}`);
        }
        return table;
    }
    tableName(src, context) {
        return this.#getTable(src, context).tableName;
    }
    columnName(table, src, ctx) {
        const dst = this.#getTable(table, ctx).columns[src];
        if (!dst) {
            throw new Error(`unknown column "${src}" of "${table}" table ${!ctx ? '' : `in ${JSON.stringify(ctx)}`}`);
        }
        return dst;
    }
    row(table, row) {
        const dest = this.#getTable(table);
        const { allColumnsSame, columns } = dest;
        if (allColumnsSame) {
            return row;
        }
        const clientRow = {};
        for (const col in row) {
            // Note: columns with unknown names simply pass through.
            clientRow[columns[col] ?? col] = row[col];
        }
        return clientRow;
    }
    columns(table, cols) {
        const dest = this.#getTable(table);
        const { allColumnsSame, columns } = dest;
        // Note: Columns not defined in the schema simply pass through.
        return cols === undefined || allColumnsSame
            ? cols
            : cols.map(col => columns[col] ?? col);
    }
}
//# sourceMappingURL=name-mapper.js.map