import { assert } from "../../../shared/src/asserts.js";
import { must } from "../../../shared/src/must.js";
import { Exists } from "../ivm/exists.js";
import { FanIn } from "../ivm/fan-in.js";
import { FanOut } from "../ivm/fan-out.js";
import { buildFilterPipeline, } from "../ivm/filter-operators.js";
import { Filter } from "../ivm/filter.js";
import { Join } from "../ivm/join.js";
import { Skip } from "../ivm/skip.js";
import { Take } from "../ivm/take.js";
import { createPredicate } from "./filter.js";
/**
 * Builds a pipeline from an AST. Caller must provide a delegate to create source
 * and storage interfaces as necessary.
 *
 * Usage:
 *
 * ```ts
 * class MySink implements Output {
 *   readonly #input: Input;
 *
 *   constructor(input: Input) {
 *     this.#input = input;
 *     input.setOutput(this);
 *   }
 *
 *   push(change: Change, _: Operator) {
 *     console.log(change);
 *   }
 * }
 *
 * const input = buildPipeline(ast, myDelegate);
 * const sink = new MySink(input);
 * ```
 */
export function buildPipeline(ast, delegate) {
    return buildPipelineInternal(delegate.mapAst ? delegate.mapAst(ast) : ast, delegate, '');
}
export function bindStaticParameters(ast, staticQueryParameters) {
    const visit = (node) => ({
        ...node,
        where: node.where ? bindCondition(node.where) : undefined,
        related: node.related?.map(sq => ({
            ...sq,
            subquery: visit(sq.subquery),
        })),
    });
    function bindCondition(condition) {
        if (condition.type === 'simple') {
            return {
                ...condition,
                left: bindValue(condition.left),
                right: bindValue(condition.right),
            };
        }
        if (condition.type === 'correlatedSubquery') {
            return {
                ...condition,
                related: {
                    ...condition.related,
                    subquery: visit(condition.related.subquery),
                },
            };
        }
        return {
            ...condition,
            conditions: condition.conditions.map(bindCondition),
        };
    }
    const bindValue = (value) => {
        if (isParameter(value)) {
            const anchor = must(staticQueryParameters, 'Static query params do not exist')[value.anchor];
            const resolvedValue = resolveField(anchor, value.field);
            return {
                type: 'literal',
                value: resolvedValue,
            };
        }
        return value;
    };
    return visit(ast);
}
function resolveField(anchor, field) {
    if (anchor === undefined) {
        return null;
    }
    if (Array.isArray(field)) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        return field.reduce((acc, f) => acc?.[f], anchor) ?? null;
    }
    return anchor[field] ?? null;
}
function isParameter(value) {
    return value.type === 'static';
}
function buildPipelineInternal(ast, delegate, name, partitionKey) {
    const source = delegate.getSource(ast.table);
    if (!source) {
        throw new Error(`Source not found: ${ast.table}`);
    }
    ast = uniquifyCorrelatedSubqueryConditionAliases(ast);
    const csqsFromCondition = gatherCorrelatedSubqueryQueriesFromCondition(ast.where);
    const splitEditKeys = partitionKey
        ? new Set(partitionKey)
        : new Set();
    const aliases = new Set();
    for (const csq of csqsFromCondition) {
        aliases.add(csq.subquery.alias || '');
        for (const key of csq.correlation.parentField) {
            splitEditKeys.add(key);
        }
    }
    if (ast.related) {
        for (const csq of ast.related) {
            for (const key of csq.correlation.parentField) {
                splitEditKeys.add(key);
            }
        }
    }
    const conn = source.connect(must(ast.orderBy), ast.where, splitEditKeys);
    let end = delegate.decorateInput(conn, `${name}:source(${ast.table})`);
    const { fullyAppliedFilters } = conn;
    if (ast.start) {
        end = delegate.decorateInput(new Skip(end, ast.start), `${name}:skip)`);
    }
    for (const csq of csqsFromCondition) {
        end = applyCorrelatedSubQuery(csq, delegate, end, name);
    }
    if (ast.where && !fullyAppliedFilters) {
        end = applyWhere(end, ast.where, delegate, name);
    }
    if (ast.limit !== undefined) {
        const takeName = `${name}:take`;
        end = delegate.decorateInput(new Take(end, delegate.createStorage(takeName), ast.limit, partitionKey), takeName);
    }
    if (ast.related) {
        for (const csq of ast.related) {
            end = applyCorrelatedSubQuery(csq, delegate, end, name);
        }
    }
    return end;
}
function applyWhere(input, condition, delegate, name) {
    return buildFilterPipeline(input, filterInput => applyFilter(filterInput, condition, delegate, name));
}
function applyFilter(input, condition, delegate, name) {
    switch (condition.type) {
        case 'and':
            return applyAnd(input, condition, delegate, name);
        case 'or':
            return applyOr(input, condition, delegate, name);
        case 'correlatedSubquery':
            return applyCorrelatedSubqueryCondition(input, condition, delegate, name);
        case 'simple':
            return applySimpleCondition(input, condition);
    }
}
function applyAnd(input, condition, delegate, name) {
    for (const subCondition of condition.conditions) {
        input = applyFilter(input, subCondition, delegate, name);
    }
    return input;
}
export function applyOr(input, condition, delegate, name) {
    const [subqueryConditions, otherConditions] = groupSubqueryConditions(condition);
    // if there are no subquery conditions, no fan-in / fan-out is needed
    if (subqueryConditions.length === 0) {
        return new Filter(input, createPredicate({
            type: 'or',
            conditions: otherConditions,
        }));
    }
    const fanOut = new FanOut(input);
    const branches = subqueryConditions.map(subCondition => applyFilter(fanOut, subCondition, delegate, name));
    if (otherConditions.length > 0) {
        branches.push(new Filter(fanOut, createPredicate({
            type: 'or',
            conditions: otherConditions,
        })));
    }
    const ret = new FanIn(fanOut, branches);
    fanOut.setFanIn(ret);
    return ret;
}
export function groupSubqueryConditions(condition) {
    const partitioned = [[], []];
    for (const subCondition of condition.conditions) {
        if (isNotAndDoesNotContainSubquery(subCondition)) {
            partitioned[1].push(subCondition);
        }
        else {
            partitioned[0].push(subCondition);
        }
    }
    return partitioned;
}
export function isNotAndDoesNotContainSubquery(condition) {
    if (condition.type === 'correlatedSubquery') {
        return false;
    }
    if (condition.type === 'simple') {
        return true;
    }
    return condition.conditions.every(isNotAndDoesNotContainSubquery);
}
function applySimpleCondition(input, condition) {
    return new Filter(input, createPredicate(condition));
}
function applyCorrelatedSubQuery(sq, delegate, end, name) {
    assert(sq.subquery.alias, 'Subquery must have an alias');
    const child = buildPipelineInternal(sq.subquery, delegate, `${name}.${sq.subquery.alias}`, sq.correlation.childField);
    const joinName = `${name}:join(${sq.subquery.alias})`;
    end = new Join({
        parent: end,
        child,
        storage: delegate.createStorage(joinName),
        parentKey: sq.correlation.parentField,
        childKey: sq.correlation.childField,
        relationshipName: sq.subquery.alias,
        hidden: sq.hidden ?? false,
        system: sq.system ?? 'client',
    });
    return delegate.decorateInput(end, joinName);
}
function applyCorrelatedSubqueryCondition(input, condition, delegate, name) {
    assert(condition.op === 'EXISTS' || condition.op === 'NOT EXISTS');
    const existsName = `${name}:exists(${condition.related.subquery.alias})`;
    return delegate.decorateFilterInput(new Exists(input, delegate.createStorage(existsName), must(condition.related.subquery.alias), condition.related.correlation.parentField, condition.op), existsName);
}
function gatherCorrelatedSubqueryQueriesFromCondition(condition) {
    const csqs = [];
    const gather = (condition) => {
        if (condition.type === 'correlatedSubquery') {
            assert(condition.op === 'EXISTS' || condition.op === 'NOT EXISTS');
            csqs.push({
                ...condition.related,
                subquery: {
                    ...condition.related.subquery,
                    limit: condition.related.system === 'permissions'
                        ? PERMISSIONS_EXISTS_LIMIT
                        : EXISTS_LIMIT,
                },
            });
            return;
        }
        if (condition.type === 'and' || condition.type === 'or') {
            for (const c of condition.conditions) {
                gather(c);
            }
            return;
        }
    };
    if (condition) {
        gather(condition);
    }
    return csqs;
}
const EXISTS_LIMIT = 3;
const PERMISSIONS_EXISTS_LIMIT = 1;
export function assertOrderingIncludesPK(ordering, pk) {
    const orderingFields = ordering.map(([field]) => field);
    const missingFields = pk.filter(pkField => !orderingFields.includes(pkField));
    if (missingFields.length > 0) {
        throw new Error(`Ordering must include all primary key fields. Missing: ${missingFields.join(', ')}. ZQL automatically appends primary key fields to the ordering if they are missing 
      so a common cause of this error is a casing mismatch between Postgres and ZQL.
      E.g., "userid" vs "userID".
      You may want to add double-quotes around your Postgres column names to prevent Postgres from lower-casing them:
      https://www.postgresql.org/docs/current/sql-syntax-lexical.htm`);
    }
}
function uniquifyCorrelatedSubqueryConditionAliases(ast) {
    if (!ast.where) {
        return ast;
    }
    const { where } = ast;
    if (where.type !== 'and' && where.type !== 'or') {
        return ast;
    }
    let count = 0;
    const uniquifyCorrelatedSubquery = (csqc) => ({
        ...csqc,
        related: {
            ...csqc.related,
            subquery: {
                ...csqc.related.subquery,
                alias: (csqc.related.subquery.alias ?? '') + '_' + count++,
            },
        },
    });
    const uniquify = (cond) => {
        if (cond.type === 'simple') {
            return cond;
        }
        else if (cond.type === 'correlatedSubquery') {
            return uniquifyCorrelatedSubquery(cond);
        }
        const conditions = [];
        for (const c of cond.conditions) {
            conditions.push(uniquify(c));
        }
        return {
            type: cond.type,
            conditions,
        };
    };
    const result = {
        ...ast,
        where: uniquify(where),
    };
    return result;
}
//# sourceMappingURL=builder.js.map