import { assert } from "../../../shared/src/asserts.js";
import { applyChange } from "./view-apply-change.js";
/**
 * Implements a materialized view of the output of an operator.
 *
 * It might seem more efficient to use an immutable b-tree for the
 * materialization, but it's not so clear. Inserts in the middle are
 * asymptotically slower in an array, but can often be done with zero
 * allocations, where changes to the b-tree will often require several allocs.
 *
 * Also the plain array view is more convenient for consumers since you can dump
 * it into console to see what it is, rather than having to iterate it.
 */
export class ArrayView {
    #input;
    #listeners = new Set();
    #schema;
    #format;
    // Synthetic "root" entry that has a single "" relationship, so that we can
    // treat all changes, including the root change, generically.
    #root;
    onDestroy;
    #dirty = false;
    #complete = false;
    #updateTTL;
    constructor(input, format, queryComplete, updateTTL) {
        this.#input = input;
        this.#schema = input.getSchema();
        this.#format = format;
        this.#updateTTL = updateTTL;
        this.#root = { '': format.singular ? undefined : [] };
        input.setOutput(this);
        if (queryComplete === true) {
            this.#complete = true;
        }
        else {
            void queryComplete.then(() => {
                this.#complete = true;
                this.#fireListeners();
            });
        }
        this.#hydrate();
    }
    get data() {
        return this.#root[''];
    }
    addListener(listener) {
        assert(!this.#listeners.has(listener), 'Listener already registered');
        this.#listeners.add(listener);
        this.#fireListener(listener);
        return () => {
            this.#listeners.delete(listener);
        };
    }
    #fireListeners() {
        for (const listener of this.#listeners) {
            this.#fireListener(listener);
        }
    }
    #fireListener(listener) {
        listener(this.data, this.#complete ? 'complete' : 'unknown');
    }
    destroy() {
        this.onDestroy?.();
    }
    #hydrate() {
        this.#dirty = true;
        for (const node of this.#input.fetch({})) {
            applyChange(this.#root, { type: 'add', node }, this.#schema, '', this.#format);
        }
        this.flush();
    }
    push(change) {
        this.#dirty = true;
        applyChange(this.#root, change, this.#schema, '', this.#format);
    }
    flush() {
        if (!this.#dirty) {
            return;
        }
        this.#dirty = false;
        this.#fireListeners();
    }
    updateTTL(ttl) {
        this.#updateTTL(ttl);
    }
}
//# sourceMappingURL=array-view.js.map