import { assert } from "../../../shared/src/asserts.js";
import { stringCompare } from "../../../shared/src/string-compare.js";
import { valuesEqual } from "./data.js";
export function constraintMatchesRow(constraint, row) {
    for (const key in constraint) {
        if (!valuesEqual(row[key], constraint[key])) {
            return false;
        }
    }
    return true;
}
export function constraintMatchesPrimaryKey(constraint, primary) {
    const constraintKeys = Object.keys(constraint);
    if (constraintKeys.length !== primary.length) {
        return false;
    }
    // Primary key is always sorted
    // Constraint does not have to be sorted
    constraintKeys.sort(stringCompare);
    for (let i = 0; i < constraintKeys.length; i++) {
        if (constraintKeys[i][0] !== primary[i]) {
            return false;
        }
    }
    return true;
}
/**
 * Pulls top level `and` components out of a condition tree.
 * The resulting array of simple conditions would match a superset of
 * values that the original condition would match.
 *
 * Examples:
 * a AND b OR c
 *
 * In this case we cannot pull anything because the `or` is at the top level.
 *
 * a AND b AND c
 * We can pull all three.
 *
 * a AND (b OR c)
 * We can only pull `a`.
 */
export function pullSimpleAndComponents(condition) {
    if (condition.type === 'and') {
        return condition.conditions.flatMap(pullSimpleAndComponents);
    }
    if (condition.type === 'simple') {
        return [condition];
    }
    if (condition.type === 'or' && condition.conditions.length === 1) {
        return pullSimpleAndComponents(condition.conditions[0]);
    }
    return [];
}
/**
 * Checks if the supplied filters constitute a primary key lookup.
 * If so, returns the constraint that would be used to look up the primary key.
 * If not, returns undefined.
 */
export function primaryKeyConstraintFromFilters(condition, primary) {
    if (condition === undefined) {
        return undefined;
    }
    const conditions = pullSimpleAndComponents(condition);
    if (conditions.length === 0) {
        return undefined;
    }
    const ret = {};
    for (const subCondition of conditions) {
        if (subCondition.op === '=') {
            const column = extractColumn(subCondition);
            if (column !== undefined) {
                if (!primary.includes(column.name)) {
                    continue;
                }
                ret[column.name] = column.value;
            }
        }
    }
    if (Object.keys(ret).length !== primary.length) {
        return undefined;
    }
    return ret;
}
function extractColumn(condition) {
    if (condition.left.type === 'column') {
        assert(condition.right.type === 'literal');
        return { name: condition.left.name, value: condition.right.value };
    }
    return undefined;
}
export class SetOfConstraint {
    #data = [];
    constructor() {
        // Only used in testing
        assert(TESTING);
    }
    #indexOf(value) {
        return this.#data.findIndex(v => constraintEquals(v, value));
    }
    has(value) {
        return this.#indexOf(value) !== -1;
    }
    add(value) {
        if (!this.has(value)) {
            this.#data.push(value);
        }
        return this;
    }
}
function constraintEquals(a, b) {
    const aEntries = Object.entries(a);
    const bEntries = Object.entries(b);
    if (aEntries.length !== bEntries.length) {
        return false;
    }
    for (let i = 0; i < aEntries.length; i++) {
        if (aEntries[i][0] !== bEntries[i][0] ||
            !valuesEqual(aEntries[i][1], bEntries[i][1])) {
            return false;
        }
    }
    return true;
}
//# sourceMappingURL=constraint.js.map