import { assert } from "../../../shared/src/asserts.js";
import { must } from "../../../shared/src/must.js";
import {} from "./data.js";
import { throwFilterOutput, } from "./filter-operators.js";
/**
 * The FanIn operator merges multiple streams into one.
 * It eliminates duplicates and must be paired with a fan-out operator
 * somewhere upstream of the fan-in.
 *
 *  issue
 *    |
 * fan-out
 * /      \
 * a      b
 *  \    /
 * fan-in
 *   |
 */
export class FanIn {
    #inputs;
    #schema;
    #output = throwFilterOutput;
    #accumulatedPushes = [];
    constructor(fanOut, inputs) {
        this.#inputs = inputs;
        this.#schema = fanOut.getSchema();
        for (const input of inputs) {
            input.setFilterOutput(this);
            assert(this.#schema === input.getSchema(), `Schema mismatch in fan-in`);
        }
    }
    setFilterOutput(output) {
        this.#output = output;
    }
    destroy() {
        for (const input of this.#inputs) {
            input.destroy();
        }
    }
    getSchema() {
        return this.#schema;
    }
    filter(node, cleanup) {
        return this.#output.filter(node, cleanup);
    }
    push(change) {
        this.#accumulatedPushes.push(change);
    }
    fanOutDonePushingToAllBranches(fanOutChangeType) {
        if (this.#inputs.length === 0) {
            assert(this.#accumulatedPushes.length === 0, 'If there are no inputs then fan-in should not receive any pushes.');
            return;
        }
        if (this.#accumulatedPushes.length === 0) {
            // It is possible for no forks to pass along the push.
            // E.g., if no filters match in any fork.
            return;
        }
        // collapse down to a single change per type
        const candidatesToPush = new Map();
        for (const change of this.#accumulatedPushes) {
            if (fanOutChangeType === 'child' && change.type !== 'child') {
                assert(candidatesToPush.has(change.type) === false, () => `Fan-in:child expected at most one ${change.type} when fan-out is of type child`);
            }
            candidatesToPush.set(change.type, change);
        }
        this.#accumulatedPushes = [];
        const types = [...candidatesToPush.keys()];
        /**
         * Based on the received `fanOutChangeType` only certain output types are valid.
         *
         * - remove must result in all removes
         * - add must result in all adds
         * - edit must result in add or removes or edits
         * - child must result in a single add or single remove or many child changes
         */
        switch (fanOutChangeType) {
            case 'remove':
                assert(types.length === 1 && types[0] === 'remove', 'Fan-in:remove expected all removes');
                this.#output.push(must(candidatesToPush.get('remove')));
                return;
            case 'add':
                assert(types.length === 1 && types[0] === 'add', 'Fan-in:add expected all adds');
                this.#output.push(must(candidatesToPush.get('add')));
                return;
            case 'edit': {
                assert(types.every(type => type === 'add' || type === 'remove' || type === 'edit'), 'Fan-in:edit expected all adds, removes, or edits');
                const addChange = candidatesToPush.get('add');
                const removeChange = candidatesToPush.get('remove');
                const editChange = candidatesToPush.get('edit');
                // If an `edit` is present, it supersedes `add` and `remove`
                // as it semantically represents both.
                if (editChange) {
                    this.#output.push(editChange);
                    return;
                }
                // If `edit` didn't make it through but both `add` and `remove` did,
                // convert back to an edit.
                //
                // When can this happen?
                //
                //  EDIT old: a=1, new: a=2
                //            |
                //          FanOut
                //          /    \
                //         a=1   a=2
                //          |     |
                //        remove  add
                //          \     /
                //           FanIn
                //
                // The left filter converts the edit into a remove.
                // The right filter converts the edit into an add.
                if (addChange && removeChange) {
                    this.#output.push({
                        type: 'edit',
                        node: addChange.node,
                        oldNode: removeChange.node,
                    });
                    return;
                }
                this.#output.push(must(addChange ?? removeChange));
                return;
            }
            case 'child': {
                assert(types.every(type => type === 'add' || // exists can change child to add or remove
                    type === 'remove' || // exists can change child to add or remove
                    type === 'child'), 'Fan-in:child expected all adds, removes, or children');
                assert(types.length <= 2, 'Fan-in:child expected at most 2 types on a child change from fan-out');
                // If any branch preserved the original child change, that takes precedence over all other changes.
                const childChange = candidatesToPush.get('child');
                if (childChange) {
                    this.#output.push(childChange);
                    return;
                }
                const addChange = candidatesToPush.get('add');
                const removeChange = candidatesToPush.get('remove');
                assert(addChange === undefined || removeChange === undefined, 'Fan-in:child expected either add or remove, not both');
                this.#output.push(must(addChange ?? removeChange));
                return;
            }
            default:
                fanOutChangeType;
        }
    }
}
//# sourceMappingURL=fan-in.js.map