/* eslint-disable @typescript-eslint/naming-convention */
/* eslint-disable @typescript-eslint/no-explicit-any */
import { resolver } from '@rocicorp/resolver';
import { assert } from "../../../shared/src/asserts.js";
import { hashOfAST } from "../../../zero-protocol/src/ast-hash.js";
import { isOneHop, isTwoHop, } from "../../../zero-schema/src/table-schema.js";
import { buildPipeline } from "../builder/builder.js";
import { ArrayView } from "../ivm/array-view.js";
import { assertNoNotExists } from "./assert-no-not-exists.js";
import { and, cmp, ExpressionBuilder, simplifyCondition, } from "./expression.js";
import {} from "./query.js";
import { DEFAULT_TTL } from "./ttl.js";
const astSymbol = Symbol();
export function ast(query) {
    return query[astSymbol];
}
export function newQuery(delegate, schema, table) {
    return new QueryImpl(delegate, schema, table, { table }, defaultFormat);
}
export function staticParam(anchorClass, field) {
    return {
        type: 'static',
        anchor: anchorClass,
        // for backwards compatibility
        field: field.length === 1 ? field[0] : field,
    };
}
export const SUBQ_PREFIX = 'zsubq_';
export const defaultFormat = { singular: false, relationships: {} };
export const newQuerySymbol = Symbol();
export class AbstractQuery {
    #schema;
    #tableName;
    #ast;
    format;
    #hash = '';
    #system;
    constructor(schema, tableName, ast, format, system) {
        this.#schema = schema;
        this.#tableName = tableName;
        this.#ast = ast;
        this.format = format;
        this.#system = system;
    }
    get [astSymbol]() {
        return this.#ast;
    }
    hash() {
        if (!this.#hash) {
            this.#hash = hashOfAST(this._completeAst());
        }
        return this.#hash;
    }
    one = () => this[newQuerySymbol](this.#schema, this.#tableName, {
        ...this.#ast,
        limit: 1,
    }, {
        ...this.format,
        singular: true,
    });
    whereExists = (relationship, cb) => this.where(({ exists }) => exists(relationship, cb));
    related = (relationship, cb) => {
        if (relationship.startsWith(SUBQ_PREFIX)) {
            throw new Error(`Relationship names may not start with "${SUBQ_PREFIX}". That is a reserved prefix.`);
        }
        cb = cb ?? (q => q);
        const related = this.#schema.relationships[this.#tableName][relationship];
        assert(related, 'Invalid relationship');
        if (isOneHop(related)) {
            const { destSchema, destField, sourceField, cardinality } = related[0];
            let q = this[newQuerySymbol](this.#schema, destSchema, {
                table: destSchema,
                alias: relationship,
            }, {
                relationships: {},
                singular: cardinality === 'one',
            });
            if (cardinality === 'one') {
                q = q.one();
            }
            const sq = cb(q);
            assert(isCompoundKey(sourceField), 'The source of a relationship must specify at last 1 field');
            assert(isCompoundKey(destField), 'The destination of a relationship must specify at last 1 field');
            assert(sourceField.length === destField.length, 'The source and destination of a relationship must have the same number of fields');
            return this[newQuerySymbol](this.#schema, this.#tableName, {
                ...this.#ast,
                related: [
                    ...(this.#ast.related ?? []),
                    {
                        system: this.#system,
                        correlation: {
                            parentField: sourceField,
                            childField: destField,
                        },
                        subquery: addPrimaryKeysToAst(this.#schema.tables[destSchema], sq.#ast),
                    },
                ],
            }, {
                ...this.format,
                relationships: {
                    ...this.format.relationships,
                    [relationship]: sq.format,
                },
            });
        }
        if (isTwoHop(related)) {
            const [firstRelation, secondRelation] = related;
            const { destSchema } = secondRelation;
            const junctionSchema = firstRelation.destSchema;
            const sq = cb(this[newQuerySymbol](this.#schema, destSchema, {
                table: destSchema,
                alias: relationship,
            }, {
                relationships: {},
                singular: secondRelation.cardinality === 'one',
            }));
            assert(isCompoundKey(firstRelation.sourceField), 'Invalid relationship');
            assert(isCompoundKey(firstRelation.destField), 'Invalid relationship');
            assert(isCompoundKey(secondRelation.sourceField), 'Invalid relationship');
            assert(isCompoundKey(secondRelation.destField), 'Invalid relationship');
            return this[newQuerySymbol](this.#schema, this.#tableName, {
                ...this.#ast,
                related: [
                    ...(this.#ast.related ?? []),
                    {
                        system: this.#system,
                        correlation: {
                            parentField: firstRelation.sourceField,
                            childField: firstRelation.destField,
                        },
                        hidden: true,
                        subquery: {
                            table: junctionSchema,
                            alias: relationship,
                            orderBy: addPrimaryKeys(this.#schema.tables[junctionSchema], undefined),
                            related: [
                                {
                                    system: this.#system,
                                    correlation: {
                                        parentField: secondRelation.sourceField,
                                        childField: secondRelation.destField,
                                    },
                                    subquery: addPrimaryKeysToAst(this.#schema.tables[destSchema], sq.#ast),
                                },
                            ],
                        },
                    },
                ],
            }, {
                ...this.format,
                relationships: {
                    ...this.format.relationships,
                    [relationship]: sq.format,
                },
            });
        }
        throw new Error(`Invalid relationship ${relationship}`);
    };
    where = (fieldOrExpressionFactory, opOrValue, value) => {
        let cond;
        if (typeof fieldOrExpressionFactory === 'function') {
            cond = fieldOrExpressionFactory(new ExpressionBuilder(this._exists));
        }
        else {
            assert(opOrValue !== undefined, 'Invalid condition');
            cond = cmp(fieldOrExpressionFactory, opOrValue, value);
        }
        const existingWhere = this.#ast.where;
        if (existingWhere) {
            cond = and(existingWhere, cond);
        }
        const where = simplifyCondition(cond);
        if (this.#system === 'client') {
            // We need to do this after the DNF since the DNF conversion might change
            // an EXISTS to a NOT EXISTS condition (and vice versa).
            assertNoNotExists(where);
        }
        return this[newQuerySymbol](this.#schema, this.#tableName, {
            ...this.#ast,
            where,
        }, this.format);
    };
    start = (row, opts) => this[newQuerySymbol](this.#schema, this.#tableName, {
        ...this.#ast,
        start: {
            row,
            exclusive: !opts?.inclusive,
        },
    }, this.format);
    limit = (limit) => {
        if (limit < 0) {
            throw new Error('Limit must be non-negative');
        }
        if ((limit | 0) !== limit) {
            throw new Error('Limit must be an integer');
        }
        return this[newQuerySymbol](this.#schema, this.#tableName, {
            ...this.#ast,
            limit,
        }, this.format);
    };
    orderBy = (field, direction) => this[newQuerySymbol](this.#schema, this.#tableName, {
        ...this.#ast,
        orderBy: [...(this.#ast.orderBy ?? []), [field, direction]],
    }, this.format);
    _exists = (relationship, cb = q => q) => {
        const related = this.#schema.relationships[this.#tableName][relationship];
        assert(related, 'Invalid relationship');
        if (isOneHop(related)) {
            const { destSchema, sourceField, destField } = related[0];
            assert(isCompoundKey(sourceField), 'Invalid relationship');
            assert(isCompoundKey(destField), 'Invalid relationship');
            const sq = cb(this[newQuerySymbol](this.#schema, destSchema, {
                table: destSchema,
                alias: `${SUBQ_PREFIX}${relationship}`,
            }, defaultFormat));
            return {
                type: 'correlatedSubquery',
                related: {
                    system: this.#system,
                    correlation: {
                        parentField: sourceField,
                        childField: destField,
                    },
                    subquery: addPrimaryKeysToAst(this.#schema.tables[destSchema], sq.#ast),
                },
                op: 'EXISTS',
            };
        }
        if (isTwoHop(related)) {
            const [firstRelation, secondRelation] = related;
            assert(isCompoundKey(firstRelation.sourceField), 'Invalid relationship');
            assert(isCompoundKey(firstRelation.destField), 'Invalid relationship');
            assert(isCompoundKey(secondRelation.sourceField), 'Invalid relationship');
            assert(isCompoundKey(secondRelation.destField), 'Invalid relationship');
            const { destSchema } = secondRelation;
            const junctionSchema = firstRelation.destSchema;
            const queryToDest = cb(this[newQuerySymbol](this.#schema, destSchema, {
                table: destSchema,
                alias: `${SUBQ_PREFIX}${relationship}`,
            }, defaultFormat));
            return {
                type: 'correlatedSubquery',
                related: {
                    system: this.#system,
                    correlation: {
                        parentField: firstRelation.sourceField,
                        childField: firstRelation.destField,
                    },
                    subquery: {
                        table: junctionSchema,
                        alias: `${SUBQ_PREFIX}${relationship}`,
                        orderBy: addPrimaryKeys(this.#schema.tables[junctionSchema], undefined),
                        where: {
                            type: 'correlatedSubquery',
                            related: {
                                system: this.#system,
                                correlation: {
                                    parentField: secondRelation.sourceField,
                                    childField: secondRelation.destField,
                                },
                                subquery: addPrimaryKeysToAst(this.#schema.tables[destSchema], queryToDest.#ast),
                            },
                            op: 'EXISTS',
                        },
                    },
                },
                op: 'EXISTS',
            };
        }
        throw new Error(`Invalid relationship ${relationship}`);
    };
    #completedAST;
    _completeAst() {
        if (!this.#completedAST) {
            const finalOrderBy = addPrimaryKeys(this.#schema.tables[this.#tableName], this.#ast.orderBy);
            if (this.#ast.start) {
                const { row } = this.#ast.start;
                const narrowedRow = {};
                for (const [field] of finalOrderBy) {
                    narrowedRow[field] = row[field];
                }
                this.#completedAST = {
                    ...this.#ast,
                    start: {
                        ...this.#ast.start,
                        row: narrowedRow,
                    },
                    orderBy: finalOrderBy,
                };
            }
            else {
                this.#completedAST = {
                    ...this.#ast,
                    orderBy: addPrimaryKeys(this.#schema.tables[this.#tableName], this.#ast.orderBy),
                };
            }
        }
        return this.#completedAST;
    }
    then(onFulfilled, onRejected) {
        return this.run().then(onFulfilled, onRejected);
    }
}
const completedAstSymbol = Symbol();
export function completedAST(q) {
    return q[completedAstSymbol];
}
export class QueryImpl extends AbstractQuery {
    #delegate;
    #system;
    constructor(delegate, schema, tableName, ast, format, system = 'client') {
        super(schema, tableName, ast, format, system);
        this.#system = system;
        this.#delegate = delegate;
    }
    get [completedAstSymbol]() {
        return this._completeAst();
    }
    [newQuerySymbol](schema, tableName, ast, format) {
        return new QueryImpl(this.#delegate, schema, tableName, ast, format, this.#system);
    }
    materialize(factoryOrTTL, ttl = DEFAULT_TTL) {
        const t0 = Date.now();
        let factory;
        if (typeof factoryOrTTL === 'function') {
            factory = factoryOrTTL;
        }
        else {
            ttl = factoryOrTTL ?? DEFAULT_TTL;
        }
        const ast = this._completeAst();
        const queryCompleteResolver = resolver();
        let queryComplete = this.#delegate.defaultQueryComplete;
        const removeServerQuery = this.#delegate.addServerQuery(ast, ttl, got => {
            if (got) {
                const t1 = Date.now();
                this.#delegate.onQueryMaterialized(this.hash(), ast, t1 - t0);
                queryComplete = true;
                queryCompleteResolver.resolve(true);
            }
        });
        const updateTTL = (newTTL) => {
            this.#delegate.updateServerQuery(ast, newTTL);
        };
        const input = buildPipeline(ast, this.#delegate);
        let removeCommitObserver;
        const onDestroy = () => {
            input.destroy();
            removeCommitObserver?.();
            removeServerQuery();
        };
        const view = this.#delegate.batchViewUpdates(() => (factory ?? arrayViewFactory)(this, input, this.format, onDestroy, cb => {
            removeCommitObserver = this.#delegate.onTransactionCommit(cb);
        }, queryComplete || queryCompleteResolver.promise, updateTTL));
        return view;
    }
    run(options) {
        this.#delegate.assertValidRunOptions(options);
        const v = this.materialize();
        if (options?.type === 'complete') {
            return new Promise(resolve => {
                v.addListener((data, type) => {
                    if (type === 'complete') {
                        v.destroy();
                        resolve(data);
                    }
                });
            });
        }
        options?.type;
        const ret = v.data;
        v.destroy();
        return Promise.resolve(ret);
    }
    preload(options) {
        const { resolve, promise: complete } = resolver();
        const ast = this._completeAst();
        const unsub = this.#delegate.addServerQuery(ast, options?.ttl ?? DEFAULT_TTL, got => {
            if (got) {
                resolve();
            }
        });
        return {
            cleanup: unsub,
            complete,
        };
    }
}
function addPrimaryKeys(schema, orderBy) {
    orderBy = orderBy ?? [];
    const { primaryKey } = schema;
    const primaryKeysToAdd = new Set(primaryKey);
    for (const [field] of orderBy) {
        primaryKeysToAdd.delete(field);
    }
    if (primaryKeysToAdd.size === 0) {
        return orderBy;
    }
    return [
        ...orderBy,
        ...[...primaryKeysToAdd].map(key => [key, 'asc']),
    ];
}
function addPrimaryKeysToAst(schema, ast) {
    return {
        ...ast,
        orderBy: addPrimaryKeys(schema, ast.orderBy),
    };
}
function arrayViewFactory(_query, input, format, onDestroy, onTransactionCommit, queryComplete, updateTTL) {
    const v = new ArrayView(input, format, queryComplete, updateTTL);
    v.onDestroy = onDestroy;
    onTransactionCommit(() => {
        v.flush();
    });
    return v;
}
function isCompoundKey(field) {
    return Array.isArray(field) && field.length >= 1;
}
//# sourceMappingURL=query-impl.js.map