import type { SQLQuery } from '@databases/sql';
import type { LogContext } from '@rocicorp/logger';
import type { LogConfig } from '../../otel/src/log-options.ts';
import type { Condition, Ordering } from '../../zero-protocol/src/ast.ts';
import type { Row } from '../../zero-protocol/src/data.ts';
import type { SchemaValue, ValueType } from '../../zero-schema/src/table-schema.ts';
import { type NoSubqueryCondition } from '../../zql/src/builder/filter.ts';
import type { Source, SourceChange, SourceChangeSet, SourceInput } from '../../zql/src/ivm/source.ts';
import { Database } from './db.ts';
/**
 * A source that is backed by a SQLite table.
 *
 * Values are written to the backing table _after_ being vended by the source.
 *
 * This ordering of events is to ensure self joins function properly. That is,
 * we can't reveal a value to an output before it has been pushed to that output.
 *
 * The code is fairly straightforward except for:
 * 1. Dealing with a `fetch` that has a basis of `before`.
 * 2. Dealing with compound orders that have differing directions (a ASC, b DESC, c ASC)
 *
 * See comments in relevant functions for more details.
 */
export declare class TableSource implements Source {
    #private;
    constructor(logContext: LogContext, logConfig: LogConfig, clientGroupID: string, db: Database, tableName: string, columns: Record<string, SchemaValue>, primaryKey: readonly [string, ...string[]]);
    get table(): string;
    /**
     * Sets the db (snapshot) to use, to facilitate the Snapshotter leapfrog
     * algorithm for concurrent traversal of historic timelines.
     */
    setDB(db: Database): void;
    connect(sort: Ordering, filters?: Condition | undefined, splitEditKeys?: Set<string> | undefined): SourceInput;
    toSQLiteRow(row: Row): Row;
    push(change: SourceChange | SourceChangeSet): void;
    genPush(change: SourceChange | SourceChangeSet): Generator<undefined, void, unknown>;
    /**
     * Retrieves a row from the backing DB by a unique key, or `undefined` if such a
     * row does not exist. This is not used in the IVM pipeline but is useful
     * for retrieving data that is consistent with the state (and type
     * semantics) of the pipeline. Note that this key may not necessarily correspond
     * to the `primaryKey` with which this TableSource.
     */
    getRow(rowKey: Row): Row | undefined;
}
/**
 * This applies all filters present in the AST for a query to the source.
 * This will work until subquery filters are added
 * at which point either:
 * a. we move filters to connect
 * b. we do the transform of removing subquery filters from filters while
 *    preserving the meaning of the filters.
 *
 * https://www.notion.so/replicache/Optional-Filters-OR-1303bed895458013a26ee5aafd5725d2
 */
export declare function filtersToSQL(filters: NoSubqueryCondition): SQLQuery;
export declare function toSQLiteTypes(columns: readonly string[], row: Row, columnTypes: Record<string, SchemaValue>): readonly unknown[];
export declare function toSQLiteTypeName(type: ValueType): "INTEGER" | "REAL" | "TEXT" | "NULL";
export declare function fromSQLiteTypes(valueTypes: Record<string, SchemaValue>, row: Row): Row;
export declare class UnsupportedValueError extends Error {
    constructor(msg: string);
}
//# sourceMappingURL=table-source.d.ts.map